/*
 * gpl_math.c --
 *
 *	Some mathematical functions for GPLib library.
 *
 * Copyright and warranties:
 *	A good part of this file comes from "nonc.c" in Yorick (version 1.4)
 *	by David Munro, so I reproduce his copyright here:
 *	  Copyright (c) 1994.  The Regents of the University of California.
 *	                       All rights reserved.
 *	
 *	All modifications (float/double types, ...) are from me:
 *	  Copyright (c) 1999, Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef _GPL_MATH_C
#define _GPL_MATH_C 1

#include "gpl_math.h"

/* Define HAVE_SINF if you have single precision math functions. */
#ifdef HAVE_SINF
# define TTYPE        float /* C-type for temporary reals. */
#else
# define TTYPE        double /* C-type for temporary reals. */
#endif

/*--------------------------------------------------------------------------*/

#define ISUFFIX i

#define RTYPE   float
#define CTYPE   gpl_fcomplex_t
#define ZERO    0.0F
#define ONE     1.0F
#define TWO     2.0F
#define THREE   3.0F
#define HALF    0.5F
#define RSUFFIX f
#define CSUFFIX c
#define GPL_HYPOT gpl_hypotf
#define HYPOT   gpl_hypotf
#define EXP     expf
#define LOG     logf
#define LOG10   log10f
#define LOG2    log2f
#define SQRT    sqrtf
#define CBRT    cbrtf
#define SIN     sinf
#define COS     cosf
#define TAN     tanf
#define ASIN    asinf
#define ACOS    acosf
#define ATAN    atanf
#define ATAN2   atan2f
#define SINH    sinhf
#define COSH    coshf
#define TANH    tanhf
#define SQRTZ   gpl_sqrtc
#define EXPZ    gpl_expc
#define LOGZ    gpl_logc
#define POWZZ   gpl_powcc
#define POWII   gpl_powii
#define POWRI   gpl_powfi
#define POWZI   gpl_powci
#define SINZ    gpl_sinc
#define COSZ    gpl_cosc
#define TANZ    gpl_tanc
#define ASINZ   gpl_asinc
#define ACOSZ   gpl_acosc
#define ATANZ   gpl_atanc
#define SINHZ   gpl_sinhc
#define COSHZ   gpl_coshc
#define TANHZ   gpl_tanhc
#define SIGNZ   gpl_signc
#define SQUAREZ gpl_squarec
#define CUBEZ   gpl_cubec
#define MULTZ   multc
#define RECIPZ  recipc
#define SIDEZ   sidec
#include __FILE__

#define RTYPE   double
#define CTYPE   gpl_dcomplex_t
#define ZERO    0.0
#define ONE     1.0
#define TWO     2.0
#define THREE   3.0
#define HALF    0.5
#define RSUFFIX d
#define CSUFFIX z
#define GPL_HYPOT gpl_hypotd
#ifdef HAVE_HYPOT
# define HYPOT  hypot
#else
# define HYPOT  GPL_HYPOT
#endif
#define EXP     exp
#define LOG     log
#define LOG10   log10
#define LOG2    log2
#define SQRT    sqrt
#define CBRT    cbrt
#define SIN     sin
#define COS     cos
#define TAN     tan
#define ASIN    asin
#define ACOS    acos
#define ATAN    atan
#define ATAN2   atan2
#define SINH    sinh
#define COSH    cosh
#define TANH    tanh
#define SQRTZ   gpl_sqrtz
#define EXPZ    gpl_expz
#define LOGZ    gpl_logz
#define POWZZ   gpl_powzz
#define POWRI   gpl_powdi
#define POWZI   gpl_powzi
#define SINZ    gpl_sinz
#define COSZ    gpl_cosz
#define TANZ    gpl_tanz
#define ASINZ   gpl_asinz
#define ACOSZ   gpl_acosz
#define ATANZ   gpl_atanz
#define SINHZ   gpl_sinhz
#define COSHZ   gpl_coshz
#define TANHZ   gpl_tanhz
#define SIGNZ   gpl_signz
#define SQUAREZ gpl_squarez
#define CUBEZ   gpl_cubez
#define MULTZ   multz
#define RECIPZ  recipz
#define SIDEZ   sidez
#include __FILE__

#else /* _GPL_MATH_C defined -------------------------------------------------*/

/* Exported functions defined in this file: */
extern void SQRTZ(RTYPE z[2], RTYPE x[2]);
extern void EXPZ(RTYPE z[2], RTYPE x[2]);
extern void LOGZ(RTYPE z[2], RTYPE x[2]);
extern void POWZZ(RTYPE z[2], RTYPE x[2], RTYPE y[2]);
extern int POWII(int b, int p);
extern RTYPE POWRI(RTYPE b, int p);
extern void POWZI(RTYPE z[2], RTYPE b[2], int p);
extern void SQUAREZ(RTYPE z[2], RTYPE x[2]);
extern void CUBEZ(RTYPE z[2], RTYPE x[2]);

extern void SINZ(RTYPE z[2], RTYPE x[2]);
extern void COSZ(RTYPE z[2], RTYPE x[2]);
extern void TANZ(RTYPE z[2], RTYPE x[2]);
extern void ASINZ(RTYPE z[2], RTYPE x[2]);
extern void ACOSZ(RTYPE z[2], RTYPE x[2]);
extern void ATANZ(RTYPE z[2], RTYPE x[2]);
extern void SINHZ(RTYPE z[2], RTYPE x[2]);
extern void COSHZ(RTYPE z[2], RTYPE x[2]);
extern void TANHZ(RTYPE z[2], RTYPE x[2]);
extern void SIGNZ(RTYPE z[2], RTYPE x[2]);

extern RTYPE GPL_HYPOT(RTYPE x, RTYPE y);

/* Private functions defined in this file: */
static void MULTZ(RTYPE z[2], RTYPE x[2]);
static void RECIPZ(RTYPE z[2]);
static void SIDEZ(RTYPE z[2], RTYPE x[2]);

/*---------------------------------------------------------------------------*/

/* hypot(x,y) is sqrt(x*x+y*y)
   The problem is that without some care, the domain of x and y will be only
   half (square root) of the complete range of floating point numbers.  */

RTYPE GPL_HYPOT(RTYPE x, RTYPE y)
{
  if (x<ZERO) x= -x;
  if (y<ZERO) y= -y;
  if (x<y) { x/= y; return y*SQRT(ONE+x*x); }
  if (x)   { y/= x; return x*SQRT(ONE+y*y); }
  return ZERO;
}

/*---------------------------------------------------------------------------*/

void SQRTZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE x0= HALF*x[0];
  register TTYPE x1= HALF*x[1];
  register TTYPE r= x1? HYPOT(x0, x1) : (x0>ZERO? x0 : -x0);
  if (x0 > ZERO) {
    z[0]= SQRT(r+x0);
    z[1]= x1/z[0];
  } else if ( (r-= x0) ) {
    z[1]= x1>=ZERO? SQRT(r) : -SQRT(r);
    z[0]= x1/z[1];
  } else {
    z[0]= z[1]= ZERO;
  }
}

void EXPZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE r= EXP(x[0]);
  z[0]= r*COS(x[1]);
  z[1]= r*SIN(x[1]);
}

void LOGZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE x0= x[0];
  z[0]= LOG(HYPOT(x0, x[1]));
  z[1]= ATAN2(x[1], x0);
}

void POWZZ(RTYPE z[2], RTYPE x[2], RTYPE y[2])
{
  RTYPE /* not TTYPE! */ logx[2], z0;
  LOGZ(logx, x);
  z0=   y[0]*logx[0] - y[1]*logx[1];
  z[1]= y[0]*logx[1] + y[1]*logx[0];
  z[0]= z0;
  EXPZ(z, z);
}

#ifdef POWII
int POWII(int b, int p)
{
  if (p==0) return b? 1 : 1/POWII(0,1);
  else if (p==1) return b;
  else if (p==2) return b*b;
  else if (p==3) return b*b*b;
  else if (p>0) {
    register int c, q= (p&3);
    p>>= 2;
    if (p==1) c= b;
    else if (p==2) c= b*b;
    else if (p==3) c= b*b*b;
    else c= POWII(b, p);      /* recurse for large powers (p>15) */
    c*= c;
    if (q&2) c*= b;
    if (q&1) return c*c*b;
    else return c*c;
  } else /* (p<0) */ {
    if (b==1) return 1;
    else if (b==-1) return (p&1)? -1 : 1; /* assumes 2s complement */
    else return 0;  /* incorrect if b==0... */
  }
}
#endif

RTYPE POWRI(RTYPE b, int p)
{
  if (p==0) return b? ONE : ONE/POWRI(ZERO, 1);
  else if (p==1) return b;
  else if (p==2) return b*b;
  else if (p==3) return b*b*b;
  else if (p>0) {
    register RTYPE c;
    register int q= (p&3);
    p>>= 2;
    if (p==1) c= b;
    else if (p==2) c= b*b;
    else if (p==3) c= b*b*b;
    else c= POWRI(b, p);      /* recurse for large powers (p>15) */
    c*= c;
    if (q&2) c*= b;
    if (q&1) return c*c*b;
    else return c*c;
  } else /* (p<0) */ {
    if (p==-p) p+= 1;        /* p=1<<31 (or 63) is <0, but p==-p */
    return ONE/POWRI(b, -p);
  }
}

void SQUAREZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE x0= x[0];
  z[0]= x0*x0 - x[1]*x[1];
  z[1]= TWO*x0*x[1];
}

void CUBEZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE x0= x[0];
  z[0]= x0*x0*x0 - THREE*x0*x[1]*x[1];
  z[1]= THREE*x0*x0*x[1] - x[1]*x[1]*x[1];
}

static void MULTZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE x0= x[0];
  register RTYPE z0= z[0];
  z[0]= x0*z0 - x[1]*z[1];
  z[1]= x[1]*z0 + x0*z[1];
}

static void RECIPZ(RTYPE z[2])
{
  register RTYPE z0= z[0];
  register RTYPE z1= z[1];
  register RTYPE nrm;
  if ((z0>0?z0:-z0)>(z1>0?z1:-z1)) {
    nrm= z0;  z0= ONE;  z1/= nrm;  nrm*= ONE+z1*z1;
  } else {
    nrm= z1;  z0/= nrm;  z1= ONE;  nrm*= ONE+z0*z0;
  }
  z[0]= z0/nrm;
  z[1]= -z1/nrm;
}

void POWZI(RTYPE z[2], RTYPE b[2], int p)
{
  if (p==0) { z[0]= (b[0]||b[1])? ONE : ONE/POWRI(ZERO, 1L);  z[1]= ZERO; }
  else if (p==1) { z[0]= b[0];  z[1]= b[1]; }
  else if (p==2) { SQUAREZ(z, b); return; }
  else if (p==3) { CUBEZ(z, b); return; }
  else if (p>0) {
    RTYPE c[2];
    register int q= (p&3);
    p>>= 2;
    if (p==1) { c[0]= b[0];  c[1]= b[1]; }
    else if (p==2) SQUAREZ(c, b);
    else if (p==3) { CUBEZ(c, b); }
    else POWZI(c, b, p);      /* recurse for large powers (p>15) */
    SQUAREZ(c, c);
    if (q&2) MULTZ(c, b);
    SQUAREZ(c, c);
    if (q&1) MULTZ(c, b);
    z[0]= c[0];
    z[1]= c[1];
  } else /* (p<0) */ {
    POWZI(z, b, -p);
    RECIPZ(z);
  }
}

/*---------------------------------------------------------------------------*/

void SINZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE sinx= SIN(x[0]);
  register TTYPE cosx= COS(x[0]);
  z[0]= sinx*COSH(x[1]);
  z[1]= cosx*SINH(x[1]);
}

void COSZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE sinx= SIN(x[0]);
  register TTYPE cosx= COS(x[0]);
  z[0]= cosx*COSH(x[1]);
  z[1]= -sinx*SINH(x[1]);
}

void TANZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE tanh2y= TANH(2*x[1]);
  register TTYPE sech2y= SQRT(ONE-tanh2y*tanh2y); /* COSH can overflow */
  register TTYPE denom= ONE + COS(2*x[0])*sech2y;
  z[0]= SIN(2*x[0])*sech2y/denom;
  z[1]= tanh2y/denom;
}

void SINHZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE sinhx= SINH(x[0]);
  register TTYPE coshx= COSH(x[0]);
  z[0]= sinhx*COS(x[1]);
  z[1]= coshx*SIN(x[1]);
}

void COSHZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE sinhx= SINH(x[0]);
  register TTYPE coshx= COSH(x[0]);
  z[0]= coshx*COS(x[1]);
  z[1]= sinhx*SIN(x[1]);
}

void TANHZ(RTYPE z[2], RTYPE x[2])
{
  register TTYPE tanh2x= TANH(2*x[0]);
  register TTYPE sech2x= SQRT(ONE-tanh2x*tanh2x); /* COSH can overflow */
  register TTYPE denom= ONE + COS(2*x[1])*sech2x;
  z[0]= tanh2x/denom;
  z[1]= SIN(2*x[1])*sech2x/denom;
}

static void SIDEZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE x0= x[0];
  z[0]= ONE - (x0*x0 - x[1]*x[1]);
  z[1]= -TWO*x0*x[1];
  SQRTZ(z, z);    /* SQRT(1-x^2) */
}

void ASINZ(RTYPE z[2], RTYPE x[2])
{
  RTYPE side[2];
  SIDEZ(side, x);
  side[0]-= x[1];
  side[1]+= x[0];
  LOGZ(side, side);
  z[0]= side[1];
  z[1]= -side[0];
}

void ACOSZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE tmp;
  RTYPE side[2];
  SIDEZ(side, x);
  tmp= side[0];
  side[0]= x[0]-side[1];
  side[1]= x[1]+tmp;
  LOGZ(side, side);
  z[0]= side[1];
  z[1]= -side[0];
}

void ATANZ(RTYPE z[2], RTYPE x[2])
{
  RTYPE side[2];
  z[0]= -x[0];
  z[1]= ONE-x[1];
  RECIPZ(z);
  side[0]= x[0];
  side[1]= ONE+x[1];
  MULTZ(side, z);
  LOGZ(side, side);
  z[0]= -HALF*side[1];
  z[1]= HALF*side[0];
}

void SIGNZ(RTYPE z[2], RTYPE x[2])
{
  register RTYPE r= HYPOT(x[0], x[1]);
  if (r!=ZERO) {
    z[0]/= r;
    z[1]/= r;
  } else {
    z[0]= ONE;
    z[1]= ZERO;
  }
}

/*---------------------------------------------------------------------------*/

#undef RTYPE
#undef CTYPE
#undef ZERO
#undef ONE
#undef TWO
#undef THREE
#undef HALF
#undef RSUFFIX
#undef CSUFFIX
#undef HYPOT
#undef GPL_HYPOT
#undef EXP
#undef LOG
#undef LOG10
#undef LOG2
#undef SQRT
#undef CBRT
#undef SIN
#undef COS
#undef TAN
#undef ASIN
#undef ACOS
#undef ATAN
#undef ATAN2
#undef SINH
#undef COSH
#undef TANH
#undef SQRTZ
#undef EXPZ
#undef LOGZ
#undef POWZZ
#undef POWII
#undef POWRI
#undef POWZI
#undef SINZ
#undef COSZ
#undef TANZ
#undef ASINZ
#undef ACOSZ
#undef ATANZ
#undef SINHZ
#undef COSHZ
#undef TANHZ
#undef SIGNZ
#undef SQUAREZ
#undef CUBEZ
#undef MULTZ
#undef RECIPZ
#undef SIDEZ

#endif /* _GPL_MATH_C defined ------------------------------------------------*/
