/*
 * gpl_sort.c --
 *
 *	Built-in index and sort functions for TiM.  The sorting algorithm
 *	is derived from Heapsort method described in "Numerical Recipes
 *	in C" (W.H. Press, B.P. Flannery, S.A. Teukolsky and W.T. Vetterling,
 *	Cambridge University Press).
 *
 * Copyright and warranty:
 *	Copyright (c) 1995, 1996, 1997, 1999 Eric THIEBAUT.
 *
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef GPL_SORT_C
#define GPL_SORT_C 1

#include "gpl_ops0.h"

/*
 * Insert definition of low levels routines
 */

#define RTYPE gpl_byte_t
#define SORT gpl_sort_b
#include __FILE__

#define RTYPE short
#define SORT gpl_sort_s
#include __FILE__

#define RTYPE int
#define SORT gpl_sort_i
#include __FILE__

#define RTYPE long
#define SORT gpl_sort_l
#include __FILE__

#define RTYPE float
#define SORT gpl_sort_f
#include __FILE__

#define RTYPE double
#define SORT gpl_sort_d
#include __FILE__

#else /* GPL_SORT_C defined ---------------------------------------------------*/
extern int *SORT(int *dst, const RTYPE *src, int number)
{
  int i, j, k, l, m;
  RTYPE q, r, s;

  for (i=0 ; i<number ; i++) {
    dst[i]= i;
  }
  if (number==1) return dst;
  l= number/2;
  m= number-1;
  for (;;) {
    if (l) {
      q= src[(k= dst[--l])];
    } else /* l==0 */ {
      q= src[(k= dst[m])];
      dst[m]= dst[0];
      if (--m == 0) {
	dst[0]= k;
	return dst;
      }
    }

    j= (i= l)*2+1;
    while (j <= m) {
      r= src[dst[j]];
      if (j < m && r < (s= src[dst[j+1]])) {
	j++;
	r= s;
      }
      if (q < r) {
	dst[i]= dst[j];
	j+= (i= j);
      } else {
	j= m+1;
      }
    }
    dst[i]= k;
  }
  return dst;
}

/*---------------------------------------------------------------------------*/
#if 0
SUM(CTYPE *dst, const CTYPE *src, int number)
{
  RTYPE s= src[0];
  for (i=1 ; i<number ; i++) s+= src[i];
}

float/double AVG(const CTYPE *src, int number)
{
  RTYPE s= src[0];
  for (i=1 ; i<number ; i++) s+= src[i];
  return (float/double)s / (float/double)number;
}

float/double RMS(const CTYPE *src, int number)
{
  int i;
  float/double x, sx, mx;
  for (i=1 ; i<number ; i++) {
    if ((x= src[i]) < 0) x= -x;
    if (x > mx) {
      mx/= x;
      sx= sx*mx*mx + 1;
      mx= x;
    } else {
      x/= mx;
      sx+= x*x;
    }
  }
  return mx * sqrt(sx / number);
}

#endif
/*---------------------------------------------------------------------------*/

#undef SORT
#undef RTYPE
#endif /* GPL_SORT_C */
