/* 
 * tclsrv_test.c --
 *
 *	Exemple of TclSrv client.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2002 Eric THIEBAUT.
 *
 *	This file is part of TclSrv.
 *
 *	TclSrv is free  software; you can redistribute it  and/or modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	TclSrv  is distributed  in the  hope that  it will  be  useful, but
 *	WITHOUT  ANY  WARRANTY;  without   even  the  implied  warranty  of
 *	MERCHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU
 *	General Public License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with TclSrv (file "COPYING"  in the top source directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330, Boston, MA 02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */
#include <stdlib.h>
#include <stdio.h>
#include <readline.h>
#include <history.h>
#include <signal.h>
#include "tclsrv.h"

int main(argc, argv)
     int argc;
     char *argv[];
{
  tclsrv_client_t *client;
  char *prompt = "> ", *line;
  int status;
  long ms = 30000; /* 30 seconds timeout */
  
  /* For a sockect/pipe channel, the SIGPIPE signal will be received if the
     peer has closed the reading end of the channel and an attempt is made
     to write data to the peer.  Ignoring the signal causes any write
     operation to returns an error with errno set to EPIPE (which is
     handled by the TclSrv library). I therefore disable the SIGPIPE signal
     for this test program. */
  signal(SIGPIPE, SIG_IGN);

  if (argc <= 1) {
    fprintf(stderr, "usage: %s socket_name\n", argv[0]);
    exit(1);
  }
  client = tclsrv_create(argv[1]);
  if (! client) {
    fprintf(stderr, "insufficient memory or invalid socket name\n");
    return EXIT_FAILURE;
  }
  if (tclsrv_connect(client, ms) != TCLSRV_OK) {
    fprintf(stderr, "[%d] %s\n", tclsrv_status(client), tclsrv_result(client));
    tclsrv_destroy(client);
    return EXIT_FAILURE;
  }
  for (;;) {
    line = readline(prompt);
    if (line) {
      if (line[0]) add_history(line);
      status = tclsrv_send(client, ms, line);
      fprintf(stderr, "[%d] %s\n", status, tclsrv_result(client));
      free(line);
    }
  }
}
