/*
 * yeti.i --
 *
 *	Main startup file for Yeti (an extension of Yorick).
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti.i,v 1.1 2005/05/24 13:06:25 eric Exp $
 *	$Log: yeti.i,v $
 *	Revision 1.1  2005/05/24 13:06:25  eric
 *	Initial revision
 *
 */

local YETI_VERSION;
local YETI_HOME;
extern _yeti_init;
/* DOCUMENT YETI_HOME       the directory where Yeti is installed
       -or- YETI_VERSION    version of current Yeti interperter
       -or- _yeti_init;

     YETI_VERSION and  YETI_DIR are global variables predefined  by Yeti to
     store  its version  number (as  "MAJOR.MINOR.MICRO")  and installation
     directory (e.g. "/usr/local/lib/yeti-VERSION").

     The  function  _yeti_init  can  be  used  to  restore  the  values  of
     YETI_VERSION and YETI_DIR.  The  first time _yeti_init is called (this
     is automatically done  at Yeti startup), it set  the default path list
     for Yeti applications  and print out Yeti greeting  message (unless in
     batch  mode).  When called  as a  function, _yeti_init()  returns Yeti
     version as a string in the form "MAJOR.MINOR.MICRO".

     A convenient way to check if your script is parsed by Yeti is to do:

       if (is_func(_yeti_init) == 2) {
         // we are in Yeti
         ...
       } else {
         // not in Yeti
         ...
       }

   SEE ALSO: Y_LAUNCH, Y_HOME, Y_SITE, Y_VERSION,
             get_path, set_path. */
_yeti_init;

extern heapsort;
/* DOCUMENT heapsort(a)
       -or- heapsort, a;
     When called as a function, returns a vector of numberof(A) longs
     containing index values such that A(heapsort(A)) is a monotonically
     increasing vector.  When called as a subroutine, performs in-place
     sorting of elements of array A.  This function uses the heap-sort
     algorithm which may be superior to the quicksort algorithm (for
     instance for integer valued arrays).  Beware that headpsort(A) and
     sort(A) differ for multidimensional arrays.
     
   SEE ALSO: sort. */

extern set_alarm;
/* DOCUMENT set_alarm, secs, callback;
     Arrange for function CALLBACK to get called with no argument in SECS
     seconds.  CALLBACK can be specified either as a Yorick function or as
     a function name.  If CALLBACK is given as a name, the symbol is
     resolved at alarm time (after SECS seconds) this permits to (re)define
     the effective CALLBACK function before alarm expires.  When called as
     a function, the returned value is the alarm time in WALL seconds.

   SEE ALSO: set_idler. */

/*---------------------------------------------------------------------------*/
/* HASH TABLE OBJECTS */

extern h_debug;
/* DOCUMENT h_debug, object, ...
     Print out some debug information on OBJECT.

     ****************************
     *** WILL BE REMOVED SOON ***
     ****************************/

extern h_new;
/* DOCUMENT h_new();
       -or- h_new(key=value, ...);
       -or- h_new("key", value, ...);
     Returns  a new  hash table  object with  member(s) KEY  set  to VALUE.
     There may be  any number of KEY-VALUE pairs. A  particular member of a
     hash table  OBJ can be  specified as a  scalar string, i.e.  "KEY", or
     using keyword syntax,  i.e. KEY=.  The keyword syntax  is however only
     possible  if  KEY is  a  valid Yorick's  symbol  name.   VALUE can  be
     anything (even a non-array object).

     A  hash  table   object  can  be  used  to   implement  some  kind  of
     object-oriented  abstraction in  Yorick.  However,  in Yorick,  a hash
     table  must have  a simple  tree structure  -- no  loops or  rings are
     allowed (loops break Yorick's memory  manager -- beware).  You need to
     be careful not to do this as the error will not be detected.

     The difference between  a hash table object and a  list object is that
     items are retrieved by key  identifier rather than by order (by h_get,
     get_member or  dot dereferenciation).   It is possible  to dereference
     the contents of OBJ using the dot operator (as for a structure) or the
     get_member function.  For instance, it is legal to do:
       obj = h_new(x=span(-7,7,100), name="my name", op=sin, scale=33);
       plg, obj.op(obj.x), obj.x;
     but  the member  must already  exists  and there  are restrictions  to
     assignation, i.e. only contents of array members can be assigned:
       obj.name() = "some other string"; // ok
       obj.name   = "some other string"; // error
       obj.x(RANGE_OR_INDEX) = EXPR;     // ok if conformable AND member X
                                         // is not a 'fast' scalar (int,
                                         // long or double scalar)
       obj.x                 = EXPR;     // error
     and  assignation cannot therefore  change the  dimension list  or data
     type of  a hash table  member.  Redefinition/creation of a  member can
     always be performed  with the h_set function which  is the recommended
     method to set the value of a hash table member.

     Hash table objects behave differently depending how they are used:

        obj.key      - de-reference hash member
        obj("key")   - returns member named "key" in hash table object OBJ,
                       this is exactly the same as: h_get(obj, "key")
        obj()        - returns number of elements in hash table OBJ
        obj(i)       - returns i-th member in hash table OBJ; i is a scalar
                       integer and can be less or equal zero to start from
                       the last one; if the hash table is unmodified,
                       obj(i) is the same as obj(keys(i)) where
                       keys=h_keys(obj).

     For instance, to explorate the whole hash table, there are two
     possibilities:
       keys = h_keys(obj);
       n = numberof(keys);   // -or- n = obj()
       for (i=1 ; i<=n ; ++i) {
         a = obj(key(i));
         ...;
       }
     or:
       n = obj();
       for (i=1 ; i<=n ; ++i) {
         a = obj(i);
         ...;
       }
     the second form is simpler but is slower for large tables.
     
     An  important point to  remember when  using hash  table is  that hash
     members are references to their contents, i.e.
       h_set, hash, member=x;
     makes an additional  reference to array X and does  not copy the array
     although you can force that, e.g.:
       tmp = x;                   // make a copy of array X
       h_set, hash, member=tmp;   // reference copy in hash table
       tmp = [];                  // delete one reference to the copy
     Because assignation  result is  its rhs (right-hand-side),  you cannot
     do:
       h_set, hash, member=(tmp = x);   // assignation result is X
     Similarly,  unlike Yorick array  data types,  a statement  like x=hash
     does not make a copy of  the hash table, it merely makes an additional
     reference to the list.


   CAVEATS:
     In  Yorick (or  Yeti), many  objects can  be used  to  reference other
     objects: pointers, lists and hash  tables.  Since Yorick uses a simple
     reference counter to delete unused object, cyclic references (i.e.  an
     object  referencing itself  either directly  or indirectly)  result in
     objects  that  will   not  be  properly  deleted.   It   is  the  user
     reponsibility to create no cyclic  references in order to avoid memory
     leaks.   Checking a  potential (or  effective) cyclic  reference would
     require recursive  investigation of all  members of the  parent object
     and could be very time consuming.


   SEE ALSO: h_copy, h_get, h_has, h_keys, h_pop, h_set, h_stat, h_first,
             h_next, _lst, get_member. */

extern h_get;
/* DOCUMENT h_get(obj, key=);
       -or- h_get(obj, "key");
     Returns  the value  of member  KEY of  hash table  object OBJ.   If no
     member  KEY exists  in  OBJ,  nil is  returned.  h_get(OBJ, "KEY")  is
     identical to get_member(OBJ, "KEY") and also to OBJ("KEY").

   SEE ALSO h_new, get_member. */

extern h_set;
/* DOCUMENT h_set, obj, key=value, ...;
       -or- h_set, obj, "key", value, ...;
     Stores VALUE in member KEY of hash table object OBJ.  There may be any
     number of KEY-VALUE pairs.  If  called as a function, the return value
     is OBJ.

   SEE ALSO h_new, h_set_copy. */

func h_set_copy(obj, ..)
/* DOCUMENT h_set_copy, obj, key, value, ...;
     Set member KEY (a scalar string) of hash table OBJ with VALUE.  Unlike
     h_set, VALUE is duplicated if it is an array.  There may be any number
     of KEY-VALUE pairs.

   SEE ALSO h_copy, h_new, h_set. */
{
  while (more_args()) {
    key = next_arg();
    value = next_arg();
    h_set, obj, key, value;
  }
  return obj;
}

func h_copy(obj, recursively)
/* DOCUMENT h_copy(obj);
       -or- h_copy(obj, recursively);
     Effectively copy  contents of  hash table object  OBJ into a  new hash
     table that is  returned.  If argument RECURSIVELY is  true, every hash
     table object contained into OBJ  get also duplicated.  This routine is
     needed because doing CPY=OBJ, where  OBJ is a hash table object, would
     only  make a  new reference  to OBJ:  CPY and  OBJ would  be  the same
     object.

   SEE ALSO h_new, h_set, h_clone. */
{
  new = h_new();
  key_list = h_keys(obj);
  n = obj(); /* number of members */
  if (recursively) {
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      member = obj(key);
      h_set, new, key, (is_hash(member) ? h_copy(member, 1) : member);
    }
  } else {
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      member = obj(key);
      h_set, new, key, member;
    }
  }
  return new;
}

/*
 * NOTE: h_clone(obj, copy=1)           is the same as h_copy(obj)
 *       h_clone(obj, copy=1, depth=-1) is the same as h_copy(obj, 1)
 */
func h_clone(obj, copy=, depth=)
/* DOCUMENT h_clone(obj, copy=, depth=);
     Make a new  hash table with same contents as OBJ.   If keyword COPY is
     true,  a fresh  copy  is  made for  array  members.  Otherwise,  array
     members are just  referenced one more time by the  new hash table.  If
     keyword DEPTH is  non-zero, every hash table object  referenced by OBJ
     get also cloned (this is  done recursively) until level DEPTH has been
     reached  (infinite recursion  if  DEPTH is  negative).   The value  of
     keyword COPY is kept the same across the recursions.
     
   SEE ALSO h_new, h_set, h_copy. */
{
  local member;
  key_list = h_keys(obj); /* list of hash keys */
  n = obj(); /* number of members */
  new = h_new();
  if (depth) {
    --depth;
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      if (copy) member = obj(key);
      else eq_nocopy, member, obj(key);      
      h_set, new, key,
        (is_hash(member) ? h_clone(member, copy=copy, depth=depth) : member);
    }
  } else if (copy) {
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      member = obj(key);
      h_set, new, key, member;
    }
  } else {
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      h_set, new, key, obj(key);
    }
  }
  return new;
}

extern h_keys;
/* DOCUMENT h_keys(obj)
     Returns list of members of hash table object OBJ as a string vector of
     key names.  The order in which keys are returned is arbitrary.

   SEE ALSO h_new, h_first, h_next. */

extern h_has;
/* DOCUMENT h_has(obj, "key")
       -or- h_has(obj, key=)
     Returns 1 if member KEY is defined in hash table object OBJ, else 0.

   SEE ALSO h_new. */

extern h_first;
extern h_next;
/* DOCUMENT h_first(obj)
       -or- h_next(obj, key)
     Get first or next key in hash table object OBJ.  Useful to run through
     all entries in a hash table (however beware that the hash table should
     be left unchanged during the scan).  For instance:

       for (key = h_first(obj) ; key ; key = h_next(obj, key)) {
         value = obj(key);
         ...;
       }

   SEE ALSO h_new, h_keys. */

func h_info(obj, sorted)
/* DOCUMENT h_info, obj;
       -or- h_info, obj, 1;
     List contents of hashj table OBJ.  Keys are sorted if second argument
     is true.
     
   SEE ALSO: h_new, h_keys, h_first, h_next, sort. */
{
  if (sorted) {
    key_list = h_keys(obj);
    key_list = key_list(sort(key_list));
    n = obj();
    for (i=1 ; i<=n ; ++i) {
      key = key_list(i);
      write, format="%s: ", key;
      info, obj(key);
    }
  } else {
    for (key = h_first(obj) ; key ; key = h_next(obj, key)) {
      write, format="%s: ", key;
      info, obj(key);
    }
  }
}

extern h_pop;
/* DOCUMENT h_pop(obj, "key")
       -or- h_pop(obj, key=)
     Pop  member KEY  out of  hash table  object OBJ  and return  it.  When
     called  as a subroutine,  the net  result is  therefore to  delete the
     member from the hash table.

   SEE ALSO h_new, h_delete. */

func h_delete(h, ..)
/* DOCUMENT h_delete(obj, "key", ...)
     Delete members KEY, ... from hash table object OBJ and return it.  Any
     KEY arguments may be present and must be array of strings or nil.

   SEE ALSO h_new, h_pop. */
{
  local key;
  while (more_args()) {
    eq_nocopy, key, next_arg();
    n = numberof(key);
    for (i=1 ; i<=n ; ++i) h_pop, h, key(i);
  }
  return h;
}

extern h_stat;
/* DOCUMENT h_stat(obj)
     Returns an histogram of the  slot occupation in hash table object OBJ.
     The  result is  a long  integer vector  with i-th  value equal  to the
     number of slots  with (i-1) items.  Note: efficient  hash table should
     keep the number of items per slot as low as possible.

   SEE ALSO h_new. */

func h_list(obj, sorted)
/* DOCUMENT h_list(obj)
       -or- h_list(obj, sorted)
     Convert hash table object OBJ  into a list: _lst("KEY1", VALUE1, ...).
     The order  of key-value pairs  is arbitrary unless argument  SORTED is
     true in which case keys get sorted in alphabetical order.

   SEE ALSO h_new, _lst, sort. */
{
  keylist = h_keys(obj);
  n = numberof(keylist);
  if (sorted && n>1) keylist = keylist(sort(keylist)(::-1));
  list = _lst();
  for (i=1 ; i<=n ; ++i) {
    /* grow the list the fast way, adding new values to its head (adding to
       the tail would make growth an N^2 proposition, as would using the
       grow function) */
    key = keylist(i);
    list = _cat(key, h_get(obj, key), list);
  }
  return list;
}

func h_cleanup(obj, recursively)
/* DOCUMENT h_cleanup, obj, 0/1;
       -or- h_cleanup(obj, 0/1);
     Delete all void  members of hash table object OBJ  and return OBJ.  If
     the second argument is a true (non nil and non-zero) empty members get
     deleted recursively.

   SEE ALSO h_new. */
{
  local member;
  keylist = h_keys(obj);
  n = numberof(keylist);
  for (i=1 ; i<=n ; ++i) {
    key = keylist(i);
    eq_nocopy, member, h_get(obj, key);
    if (is_void(member)) h_pop, obj, key;
    else if (recursively && is_hash(member)) h_cleanup, member, recursively;
  }
  return obj;
}

func h_save_symbols(____l____, ..)
/* DOCUMENT h_save_symbols(namelist, ...);
       -or- h_save_symbols(flag);
     Return  hash  table which  references  symbols  given  in NAMELIST  or
     selected by FLAG (see symbol_names).  Of course, the symbol names will
     be used as member names in the result.

   SEE ALSO h_new, h_restore_builtin, symbol_names. */
{
  /* Attempt to use dummy symbol names in this routine to avoid clash with
     the symbols ddefined in caller's context. */
  while (more_args()) grow, ____l____, next_arg();
  if ((____s____ = structof(____l____)) != string) {
    if ((____s____!=long && ____s____!=int && ____s____!=short &&
         ____s____!=char) || dimsof(____l____)(1))
      error, "expected a list of names, or nil, or a scalar integer";
    ____l____ = symbol_names(____l____);
  }
  ____s____ = h_new();
  ____n____ = numberof(____l____);
  for (____i____=1 ; ____i____<=____n____ ; ++____i____) {
    ____k____ = ____l____(____i____);
    h_set, ____s____, ____k____, symbol_def(____k____);
  }
  return ____s____;
}

local __h_saved_builtins;
func h_restore_builtin(name) { return h_get(__h_saved_builtins, name); }
/* DOCUMENT h_restore_builtin(name);
     Get the original definition of builtin function NAME which is possible
     because,  at  startup,  all   definitions  of  builtin  functions  get
     referenced in  global hash table: __h_saved_builtins.   This is useful
     if you deleted by accident a  builtin function and want to recover it;
     for instance:
       sin = 1;
       ...
       sin = h_restore_builtin("sin");
     would restore the  definition of the sine function  that was redefined
     by the assignation.

   SEE ALSO h_new, h_save_symbols. */
if (! batch() && ! is_hash(__h_saved_builtins)) {
  __h_saved_builtins = h_save_symbols(32);
}

/*---------------------------------------------------------------------------*/
/* 1D CONVOLUTION AND "A TROUS" WAVELET TRANSFORM */

extern __yeti_convolve_f;
/* PROTOTYPE
     void yeti_convolve_f(float array dst, float array src, int stride,
                          int n, int nafter, float array ker, int w,
		          int scale, int border, float array ws); */

extern __yeti_convolve_d;
/* PROTOTYPE
     void yeti_convolve_d(double array dst, double array src, int stride,
		          int n, int nafter, double array ker, int w,
                          int scale, int border, double array ws); */

func yeti_convolve(a, which=, kernel=, scale=, border=, count=)
/* DOCUMENT ap = yeti_convolve(a)
     Convolve  array A along  its dimensions  (all by  default) by  a given
     kernel.  By default, the convolution kernel is [1,4,6,4,1]/16.0.  This
     can be  changed by using keyword  KERNEL (but the kernel  must have an
     odd number  of elements).  The following operation  is performed (with
     special  handling for the  boundaries, see  keyword BORDER)  along the
     direction(s) of interest:
     |         ____
     |         \
     |   AP(i)= \ KERNEL(j+W) * A(i + j*SCALE)
     |          /
     |         /___
     |     -W <= j <= +W
     |
     where  numberof(KERNEL)=2*W+1.  Except  for  the SCALE  factor, AP  is
     mostly  a convolution  of A  by array  KERNEL along  the  direction of
     interest.

     Keyword WHICH can be used  to specify the dimension(s) of interest; by
     default, all  dimensions get convolved.   As for indices,  elements in
     WHICH less than  1 is taken as relative to the  final dimension of the
     array.  You may specify  repeated convolution along some dimensions by
     using them several times in array WHICH (see keyword COUNT).

     Keyword BORDER can be used to set the handling of boundary conditions:
       BORDER=0  Extrapolate  missing  values  by  the left/rightmost  ones
                 (this is the default behaviour).
       BORDER=1  Extrapolate missing left  values by zero and missing right
                 values by the rightmost one.
       BORDER=2  Extrapolate  missing left  values by the  leftmost one and
                 missing right values by zero.
       BORDER=3  Extrapolate missing left/right values by zero.
       BORDER=4  Use periodic conditions.
       BORDER>4 or BORDER<0
                 Do   not   extrapolate   missing  values   but   normalize
                 convolution product  by sum  of kernel weights  taken into
                 account (assuming they are all positive).

     By  default, SCALE=1 which  corresponds to  a simple  convolution.  An
     other value can be used thanks  to keyword SCALE (e.g. for the wavelet
     "a trou" method).  The value of SCALE must be a positive integer.

     Keyword COUNT  can be used to  augment the amount  of smoothing: COUNT
     (default COUNT=1) is  the number of convolution passes.   It is better
     (i.e. faster) to use only one pass with appropriate convolution kernel
     (see keyword KERNEL).

  SEE ALSO yeti_wavelet.

  RESTRICTIONS
    1. Should use the in-place ability of the operation to limit the number
       of array copies.
    2. Complex convolution not yet  implemented (although it exists in the
       C-code). */
{
  /* Check data type of A. */
  type = structof(a);
  if (type == complex) {
    return (yeti_convolve(double(a), which=which, kernel=kernel, scale=scale,
                          border=border, count=count)
            + 1i*yeti_convolve(a.im, which=which, kernel=kernel, scale=scale,
                               border=border, count=count));
  } else if (type == double) {
    op = __yeti_convolve_d;
  } else if (type == float || type == long || type == int || type == short ||
             type == char) {
    op = __yeti_convolve_f;
    type = float;
  } else {
    error, "bad data type";
  }
  a = type(a);  /* force a private copy of A */

  /* Check dimensions of A and keyword WHICH. */
  dims = dimsof(a);
  rank = dims(1);
  if (is_void(which)) {
    which = indgen(rank);
  } else {
    which += (which <= 0)*rank;
    if (min(which) < 1 || max(which) > rank)
      error, "dimension index out of range in WHICH";
  }

  /* Check KERNEL and other keywords. */
  if (is_void(kernel)) {
    k0= type(0.375);  /* 6.0/16.0 */
    k1= type(0.25);   /* 4.0/16.0 */
    k2= type(0.0625); /* 1.0/16.0 */
    kernel= [k2, k1, k0, k1, k2];
  }
  if ((w = numberof(kernel))%2 != 1)
    error, "KERNEL must have an odd number of elements";
  if (is_void(scale)) scale = 1;
  else if (structof(scale+0)!=long || scale<=0)
    error, "bad value for keyword SCALE";
  if (is_void(border)) border = 0;
  if (is_void(count)) count = 1;

  /* Compute strides. */
  stride = array(1, rank);
  for (s=1,i=2 ; i<=rank ; ++i) stride(i) = stride(i-1)*dims(i);
  stride = stride(which);
  dims = dims(which + 1);
  nafter = numberof(a)/(dims*stride);

  /* Apply the operator along every dimensions of interest. */
  for (i=1 ; i<=numberof(which) ; ++i) {
    len = dims(i);
    for (j=1 ; j<=count ; ++j) {
      op, a, a, stride(i), len, nafter(i), kernel, (w-1)/2, scale, border,
        array(type, 2*len);
    }
  }
  return a;
}

func yeti_wavelet(a, order, which=, kernel=, border=)
/* DOCUMENT cube = yeti_wavelet(a, order)
     Compute the "a trou" wavelet transform of A.  The result is such
     that:
       CUBE(.., i) = S_i - S_(i+1)
     where:
       S_1 = A
       S_(i+1) = yeti_convolve(S_i, SCALE=2^(i-1))
     As a consequence:
       CUBE(..,sum) = A;

  SEE ALSO yeti_convolve. */
{
  if (((s=structof(order)) != long && s!=int && s!=short && s!=char) ||
      dimsof(order)(1) || order<0) {
    error, "ORDER must be a non-negative integer";
  }
  dims = dimsof(a);
  grow, dims, order+1;
  ++dims(1);
  cube = array(structof(a(1)+0.0f), dims);
  for (scale=1, i=1 ; i<=order ; ++i, scale*=2) {
    ap = a;
    a = yeti_convolve(a, which=which, kernel=kernel, scale=scale,
                      border=border);
    cube(..,i) = ap-a;
  }
  cube(..,0) = a;
  return cube;
}

extern smooth3;
/* DOCUMENT smooth3(a)
     Returns array A smoothed by a simple 3-element convolution (but for
     the edges).  In one dimension, the smoothing operation reads:
        smooth3(A)(i) = C*A(i) + D*(A(i-1) + A(i+1))
     but for the first and last element for which:
        smooth3(A)(1) = E*A(1) + D*A(2)
        smooth3(A)(n) = E*A(n) + D*A(n-1)
     where N is the length of the dimension and the coefficients are:
        C = 0.5
        D = 0.25
        E = 0.75   
     With the default value of C (see keyword C below), the smoothing
     operation is identical to:
     
        smooth3(A) = A(pcen)(zcen)             for a 1D array
        smooth3(A) = A(pcen,pcen)(zcen,zcen)   for a 2D array
        ...                                    and so on

     Keyword C can be used to specify another value for the coefficient
     C (default: C=0.5); coefficients D and E are computed as follows:
        D = 0.5*(1 - C)
        E = 0.5*(1 + C)

     The default is to smooth A along all its dimensions, but keyword WHICH
     can be used to specify the only dimension to smooth.  If WHICH is less
     or equal zero, then the smoothed dimension is the last one + WHICH.

     The smoothing operator implemented by smooth3 has the following
     properties:

     1. The smoothing operator is linear and symmetric (for any number of
        dimensions in A).  The symmetry of the smoothing operator is
        important for the computation of gradients in regularization.  For
        instance, let Y = smooth3(X) and Q be a scalar function of Y, then
        then the gradient of Q with respect to X is simply:
           DQ_DX = smooth3(DQ_DY)
        where DQ_DY is the gradient of Q with respect to Y.

     2. For a vector, A, smooth3(A)=S(,+)*A(+) where the matrix S is
        tridiagonal:

           [E D         ]
           [D C D       ]
           [  D C D     ]
           [   \ \ \    ]    where, to improve readability,
           [    \ \ \   ]    missing values are all zero.
           [     D C D  ]
           [       D C D]
           [         D E]

        You can, in principle, reverse the smoothing operation with TDsolve
        along each dimensions of smooth3(A).  Note: for a vector A, the
        operator S-I applied to A (where I is the identity matrix) is the
        finite difference 2nd derivatives of A (but for the edges).

     3. The definition of coefficients C, D and E insure that the smoothing
        operator does not change the sum of the element values of its
        argument, i.e.: sum(smooth3(A)) = sum(A).

     4. Only an array with all elements having the same value is invariant
        by the smoothing operator.  In fact "slopes" along dimensions of A
        are almost invariant, only the values along the edges are changed.
     

   KEYWORDS: c, which.
   
   SEE ALSO: TDsolve. */

/*---------------------------------------------------------------------------*/
/* GIST WINDOW */
extern window_geometry;
/* DOCUMENT window_geometry()
       -or- window_geometry(win)
     Get geometry settings of the  visible region of display window WIN (or
     current window  if WIN is nil  or not specified).   These settings are
     subject to change  each time the window get resized.   The result is a
     vector of 6 doubles:
       [DPI, ONE_PIXEL, XBIAS, YBIAS, WIDTH, HEIGHT]
     where:
       DPI = dot-per-inch of WIN
       ONE_PIXEL = pixel size in NDC units
       XBIAS = abscissa offset in NDC units
       YBIAS = ordinate offset in NDC units
       WIDTH = width of visible region in pixels
       HEIGHT = height of visible region in pixels
     Pixel coordinates (XPIX,YPIX) run  from top-left (0,0) to bottom-right
     (WIDTH-1,HEIGHT-1).  The conversion to NDC coordinates is:
       XNDC = XBIAS + XPIX*ONE_PIXEL;
       YNDC = YBIAS - YPIX*ONE_PIXEL;

     If window WIN does not exists, all output values are zero.

     Notes:
       (1) The  top/left  margin(s) used  by  Gist window  to display  some
           message are not considered as part of the "visible" region.
       (2) An  extra 0.5  pixel offset has  been added to  (XBIAS,YBIAS) to
           avoid rounding errors.

   SEE ALSO: window, current_window, viewport, limits. */

/*---------------------------------------------------------------------------*/
/* STRING ROUTINES */

extern strtrim;
func strtrimleft(s)  {return strtrim(s, 1);}
func strtrimright(s) {return strtrim(s, 2);}
/* DOCUMENT strtrim(s);
       -or- strtrim(s, flags);
       -or- strtrimleft(s);
       -or- strtrimrigth(s);
     Returns input  (array of) string(s) S without  leading and/or trailing
     blanks.  Each element  of S that is only made of  blanks, gives "" (an
     empty string).  Each element of S that is NULL gives NULL.

     Leading  spaces are  removed, if  1st bit  of FLAGS  is  set; trailing
     spaces are  removed, if 2nd  bit of FLAGS  is set, leading  spaces are
     removed.   If FLAGS  is  omitted, leading  _and_  trailing spaces  get
     removed. Alternatively, strtrimleft(S)  or strtrim(S,1) remove leading
     spaces; strtrimright(S) or strtrim(S,2) removes trailing spaces.

   SEE ALSO strlower, strupper, strlen, strpart, strtok. */

extern strlower;
extern strupper;
/* DOCUMENT strlower(s);
       -or- strupper(s);
     Returns  input (array of)  string(s) S  converted to  lower/upper case
     letters.  Each element of S that  is only made of blanks, gives "" (an
     empty string).  Each element of S that is NULL gives NULL.

   SEE ALSO strtrim, strupper, strlen, strpart, strtok. */

/*---------------------------------------------------------------------------*/
/* REGULAR EXPRESSIONS */

extern regcomp;
/* DOCUMENT regcomp(reg);
     Compile regular expression STR for further use.  REG must be a scalar
     string.

     If keyword NEWLINE  is true (non-nil and non-zero),  then newlines are
     treated specifically (i.e. as if text lines separated by newlines were
     individual strings):
       - Match-any-character operators don't match a newline.
       - A non-matching  list ([^...])  not  containing a  newline does not
         match a newline.
       - Match-beginning-of-line  operator  (^) matches  the  empty  string
         immediately after a newline,  regardless of whether keyword NOTBOL
         is true when the regular expression is used in regsub or regmatch.
       - Match-end-of-line   operator   ($)  matches    the   empty  string
         immediately before a newline, regardless of whether keyword NOTEOL
         is true when the regular expression is used.

     The default is to treat newline as any other character.

     If  keyword BASIC  is true  (non-nil and  non-zero), then  POSIX Basic
     Regular  Expression  syntax is  used;  the  default  is to  use  POSIX
     Extended Regular Expression syntax.

     If  keyword  ICASE  is   true  (non-nil  and  non-zero),  the  regular
     expression will  be case insensitive. The default  is to differentiate
     case.

     Keyword  NOSUB  can  be  set  to  a  non-zero  value  if  support  for
     subexpression addressing  of matches is  not required (see  regsub and
     regmatch).

   KEYWORDS: basic, icase, newline, nosub.

   SEE ALSO: regmatch, regsub. */

extern regmatch;
/* DOCUMENT regmatch(reg, str);
       -or- regmatch(reg, str, match0, match1, match2, match3, ...);
     Test if regular  expression REG matches string array  STR.  The result
     is an array of int's with same dimension list as STR and with elements
     set to 1/0 whether REG match or not the corresponding element in STR.

     REG can be a scalar string  which get compiled at runtime and the same
     compilation keywords  as those accepted  by regcomp can  be specified:
     BASIC,  ICASE,  NEWLINE  and/or  NOSUB.   Otherwise,  REG  must  be  a
     pre-compiled  regular  expression  (by  regcomp)  and  no  compilation
     keyword can be specified.

     Keyword START can be used to specify a starting index for the matching
     in the  input string(s).  If START is  less or equal zero,  then it is
     counted from the end of the input string(s): START=0 to start with the
     last character,  START=-1 to  start with the  2 last  characters.  The
     match  is considered  as  a  failure when  the  starting character  is
     outside  the string.   The default  is START=1  (start with  the first
     character ot every input string).

     If   keyword   NOTBOL   is    true   (non-nil   and   non-zero),   the
     match-beginning-of-line   operator   (^)   always  fails   to   match.
     Similarly, if  keyword NOTEOL is true,  the match-end-of-line operator
     ($) always fails to match.   These keywords may be used when different
     portions of a string are  matched against a regular expression and the
     beginning/end  of  the  string   should  not  be  interpreted  as  the
     beginning/end of the line (but see the keyword NEWLINE in regcomp).

     In the second  form, matching patterns get stored  in output variables
     MATCH0,  MATCH1...  MATCH0  will be  filled by  the part  of  STR that
     matched the  complete regular expression REG;  MATCHn (with n=1,2...),
     will  contain the  part of  STR  that matched  the n-th  parenthesized
     subexpression  in  REG.   If  keyword  INDICES is  true  (non-nil  and
     non-zero) the indices of matching (sub)expressions get stored into the
     MATCHn variables: MATCHn(1,..) and MATCHn(2,..) are the indices of the
     first  and last+1 characters  of the  matching (sub)expressions  -- it
     must be last+1  to allow for empty match.  If  keyword INDICE is false
     or  omitted, the  MATCHn variables  will  be string  arrays with  same
     dimension lists as STR.

   KEYWORDS: basic, icase, indices, newline, nosub, notbol, noteol.

   SEE ALSO: regcomp, , regmatch_part, regsub. */


extern regsub;
/* DOCUMENT regsub(reg, str)
       -or- regsub(reg, str, sub)
     Substitute pattern(s) in STR matching regular expression REG by string
     SUB (which is the nil string  by default). STR is an array of strings,
     and the result has the same dimension list as STR.  Each string of STR
     is matched  against regular expression REG  and, if there  is a match,
     the matching part is replaced by  SUB (which must be a scalar string).
     If SUB contains a sequence  "\n" (where n=0,...,9), then this sequence
     is replaced in  the result by the n-th  parenthesized subexpression of
     REG  ("\0"  stands for  the  whole  matching  part).  Other  backslash
     sequences "\c" (where  c is any character) in SUB  get replaced by the
     escaped character c.  Beware that Yorick's parser interprets backslash
     sequences  in literal  strings,  you  may therefore  have  to write  2
     backslashes.

     REG can be a scalar string  which get compiled at runtime and the same
     compilation keywords  as those accepted  by regcomp can  be specified:
     BASIC,  ICASE,  NEWLINE  and/or  NOSUB.   Otherwise,  REG  must  be  a
     pre-compiled  regular  expression  (by  regcomp)  and  no  compilation
     keyword can be specified.

     If keyword ALL is true  (non-nil and non-zero), then all occurences of
     REG  get substituted.   Otherwise only  the 1st  occurence of  REG get
     substituted.  Of course substitution is performed for every element of
     STR.

     If   keyword   NOTBOL   is    true   (non-nil   and   non-zero),   the
     match-beginning-of-line   operator   (^)   always  fails   to   match.
     Similarly, if  keyword NOTEOL is true,  the match-end-of-line operator
     ($) always fails to match.   These keywords may be used when different
     portions of a string are  matched against a regular expression and the
     beginning/end  of  the  string   should  not  be  interpreted  as  the
     beginning/end of the line (but see the keyword NEWLINE above).

   KEYWORDS: all, basic, icase, newline, nosub, notbol, noteol.

   SEE ALSO: regcomp, regmatch. */

func regmatch_part(s, i)
/* DOCUMENT regmatch_part(str, idx);
     Get part  of string  STR indexed by  IDX (which  should be one  of the
     outputs of  regmatch when called with keyword  INDICES=1).  The result
     is an array of strings with same dimension list as STR.  IDX must have
     one more dimension than STR, the  first dimension of IDX must be 2 and
     the other must be identical to those of STR.

   SEE ALSO regmatch. */
{
  sp = array(string, dimsof(s));
  i1 = i(1,);
  i2 = i(2,) - 1;
  i = where(i2 >= i1);
  n = numberof(i);
  for (k=1 ; k<=n ; ++k) {
    ik = i(k);
    sp(ik) = strpart(s(ik), i1(ik):i2(ik));
  }
  return sp;
}

/*---------------------------------------------------------------------------*/
/* FILE ROUTINES */

extern filepath;
/* DOCUMENT filepath(file);
     Return full path name of file(s).  Argument FILE can be either an open
     binary/text file or  an array of file names (in  the latter case tilde
     expansion  is performed and  the result  will have  the same  shape as
     input).

   SEE ALSO open. */

extern expand_path;
/* DOCUMENT expand_path(path)
     Expands directory names in string array PATH (i.e. interprets relative
     location and tilde notation).  The result is an array of strings (same
     shape as PATH) of absolute paths.

   SEE ALSO cd, get_cwd, get_home, get_path. */

extern chn_open;
/*DOCUMENT chn_open(filename, mode);
    Opens FILENAME  according to  MODE and returns  a channel  object (file
    handle).   The   channel  is   automatically  closed  when   no  longer
    referenced.  MODE  is a  string made  of two parts,  the first  part is
    mandatory and is one of:
       r    - Open text file for reading only.  The stream is positioned at
              the beginning of the file.
       r+   - Open for reading and  writing.   The stream is  positioned at
              the beginning of the file.
       w    - Truncate file to zero length or create text file for writing.
              The stream is positioned at the beginning of the file.
       w+   - Open for reading and writing.  The file is created if it does
              not  exist,  otherwise  it   is  truncated.   The  stream  is
              positioned at the beginning of the file.
       a    - Open for writing.  The file is created if it does not exist.
              The stream is positioned at the end of the file.
       a+   - Open for reading and writing.  The file is created if it does
              not exist.  The stream is positioned at the end of the file.
     After the first character of MODE,  a 'b' may be present to indicate a
     binary file  (no incidence  on Unix system).   Second part of  MODE is
     optional and is used to indicate if (de)compression is required:
       z    - Use (de)compression.
       z0-9 - Level of compression (0 means no compression, 1 means fastest
              but  least  efficent method  and  9  means  slowest but  most
              efficent method).
       z[0-9][fh]
            - Level of compression  and strategy: 'f' for  filtered data as
              in "wb6f", 'h' for Huffman only compression.

    You can use 'z' to read un-compressed file (some seek operations may be
    forbidden in this case).

 SEE ALSO: chn_read, chn_rewind, chn_seek, chn_tell, chn_write. */

extern chn_seek;
/* DOCUMENT chn_seek(chn, offset, whence)
     Reposition the position of the channel CHN at location OFFSET
     according to the directive WHENCE:
       WHENCE=0 - channel position is set to OFFSET bytes;
       WHENCE=1 - channel position is set to the current location plus
                  OFFSET bytes;
       WHENCE=2 - channel position is set to the end of the file plus
                  OFFSET bytes.
     If CHN  is related  to a compressed  file, WHENCE=2 (SEEK_END)  is not
     supported.

     The chn_seek function allows the file  offset to be set beyond the end
     of the existing end-of-file of the  file.  If data is later written at
     this point,  subsequent reads of the  data in the gap  return bytes of
     zeros (until data is actually written into the gap).

     The  return value  of chn_seek  is  the resulting  offset location  as
     measured in bytes from the beginning of the file.

     SEE ALSO chn_open, chn_rewind, chn_tell. */

extern chn_tell;
/* DOCUMENT chn_tell(chn)
     Get the current  value of the file position  indicator for the channel
     CHN.  The result is the same as with: chn_seek(chn, 0, 1);

   SEE ALSO chn_open, chn_seek. */

func chn_rewind(chn) { chn_seek, chn, 0, 2; return chn; }
/* DOCUMENT chn_rewind(chn);
     Reposition the access position of  the channel CHN at beginning of the
     file and return CHN.

   SEE ALSO chn_open, chn_seek. */

extern chn_read;
/* DOCUMENT chn_read(chn, type, dimlist, ...)
       -or- chn_read(chn, nbytes)
     Read data from channel CHN.  In the first form, an array of given type
     and dimensions get  read from CHN and is returned  to the caller; only
     basic  numerical types  can be  read: char,  short, int,  long, float,
     double or complex.   In the second form, the  routine attempts to read
     NBYTES (NBYTES >  0) characters in CHN an returns  a vector (1D array)
     of char's with the  actual number of bytes read or nil  if no data are
     available; if  the returned  array is nil  or shorter than  NBYTES, it
     usually means that end-of-file has been reached.

   SEE ALSO chn_open, chn_seek, chn_write.  */

extern chn_write;
/* DOCUMENT chn_write(chn, data, ...)
     Write array DATA to channel CHN.  There may be any number of arguments
     after DATA  to write many arrays  in a single call.   Only basic array
     types  can be  written to  CHN; in  the case  of string  arrays, every
     element  ot  the array  get  written to  CHN  without  the final  null
     character.  The  return value of the  function is the  total number of
     bytes written to CHN.

   SEE ALSO chn_open, chn_read, chn_seek. */

/*---------------------------------------------------------------------------*/
/* MATH ROUTINES */

extern sinc;
/* DOCUMENT sinc(x);
     Returns the "sampling function" of X as defined by Woodward (1953) and
     Bracewell (1999):

       sinc(x) = 1                 for x=0
                 sin(PI*x)/(PI*x)  otherwise

     Note: Some other authors may  define the sampling function without the
     PI factors in the above expression.


   REFERENCES
     Bracewell, R. "The Filtering  or Interpolating Function, sinc(x)."  In
     "The  Fourier Transform  and  Its Applications",  3rd  ed.  New  York:
     McGraw-Hill, pp. 62-64, 1999.

     Woodward, P.  M. "Probability and Information Theory with Applications
     to Radar". New York: McGraw-Hill, 1953.

  SEE ALSO: sin. */

extern round;
/* DOCUMENT round(x);
     Returns X rounded to the nearest integer (as a floating point value). */

extern arc;
/* DOCUMENT arc(x);
     Returns angle X wrapped in range q(-PI, +PI] */

/*---------------------------------------------------------------------------*/
/* SPARSE ARRAYS AND MATRIX-VECTOR MULTIPLICATION */

func sparse_make(a)
/* DOCUMENT s = sparse_make(a);
     Convert array A to sparse structure, a vector of 3 pointers:
       *S(1) = non-zero elements of A;
       *S(2) = indices of non-zero elements of A;
       *S(3) = dimension list of A as given by dimsof();

   SEE ALSO: sparse_expand, mvmult. */
{
  if (! is_array((i = where(a)))) error, "input array is zero everywhere!";
  return [&a(i), &i, &dimsof(a)];
}

func sparse_expand(s)
/* DOCUMENT a = sparse_expand(s);
     Convert sparse array S into standard Yorick's array A.

   SEE ALSO: sparse_make. */
{
  (a = array(structof(*s(1)), *s(3)))(*s(2)) = *s(1);
  return a;
}

extern _mvmult;
/* DOCUMENT y = _mvmult(a, x, 0/1);
       -or- y = _mvmult(aa, ai, ad, x, 0/1);
     Return  result  of  (generalized)  matrix-vector  multiplication.   If
     optional last argument is  non-zero, the matrix operator is transposed
     (default is  no transpose).  In the  first form, A is  the full matrix
     array.  The second  form is designed for sparse operator  A: AA is the
     vector  of non-zero  elements of  A, AI  is the  vector of  indices of
     non-zero elements (starting at 1) and DA is the vector dimsof(A).

   SEE ALSO: mvmult. */

func mvmult(a, x, tr)
/* DOCUMENT y = mvmult(a, x);
       -or- y = mvmult(a, x, 1);
     Return result of (generalized) matrix-vector multiplication.  A is the
     so-called `matrix' (a normal Yorick  array or a sparse array structure
     as returned by sparse_make) of rank NA (number of dimensions) and X is
     the so-called  `vector' of rank NX.   The matrix-vector multiplication
     is performed as if there is only  one index running over the rank of X
     and the trailing/leading NX dimensions of A.

     If optional last argument is non-nil and non-zero, the matrix operator
     is transposed:  the NX  leading dimensions  of A must  be the  same as
     those of  X and the  dimensions of the  result are the  NA-NX trailing
     dimensions of A.

     Otherwise (the default), the matrix operator is not transposed: the NX
     trailing  dimensions of  A must  be the  same as  those of  X  and the
     dimensions of the result are the NA-NX leading dimensions of A.

   RESTRICTIONS:
     Only double precision arrays are supported.

   SEE ALSO: _mvmult, sparse_make. */
{
  if (structof(a)==pointer) return _mvmult(*a(1), *a(2), *a(3), x, tr);
  return _mvmult(a, x, tr);
}

/*---------------------------------------------------------------------------*/
/* MEMORY HACKING ROUTINES */

extern mem_base;
extern mem_copy;
extern mem_peek;
/* DOCUMENT mem_base(array);
       -or- mem_copy, address, expression;
       -or- mem_peek(address, type, dimlist);
     Hacker routines  to read/write data  at given memory  location.  These
     routines allow the user to de _very_ nasty but sometimes needed things
     and do  not provide  the safety level  of ususal Yorick  routines, and
     must therefore be used with  extreme care (you've bee warned).  In all
     these routines,  ADDRESS is either a  long integer scalar  or a scalar
     pointer (e.g. &OBJECT).

     mem_base returns the address  (as a long scalar)  of the first element
              of array object ARRAY.  You can use this function if you need
              to add some offset to the address of an object, e.g. to reach
              some particular element of an array or a structure.

     mem_copy copy the contents of EXPRESSION at memory location ADDRESS.

     mem_peek returns  a new  array of data  type TYPE  and dimension  list
              DIMLIST  filled  with  memory  contents starting  at  address
              ADDRESS.

   EXAMPLE
     The following statement converts the contents of complex array Z as an
     array of doubles:
       X = mem_peek(mem_base(Z), double, 2, dimsof(Z));
     then:
       X(1,..) is Z.re
       X(2,..) is Z.im

   SEE ALSO reshape, native_byte_order. */

func native_byte_order(type)
/* DOCUMENT native_byte_order()
       -or- native_byte_order(type)
     Returns the native byte  order, one of: "LITTLE_ENDIAN", "BIG_ENDIAN",
     or  "PDP_ENDIAN".  Optional  argument  TYPE is  an  integer data  type
     (default is long).

   SEE ALSO mem_peek. */
{
  if (is_void(type)) type = long;
  size = sizeof(type);
  (carr = array(char, size))(*) = indgen(size:1:-1);
  value = mem_peek(mem_base(carr), type);
  if (size == 4) {
    if (value == 0x01020304) {
      return "LITTLE_ENDIAN";
    } else if (value == 0x04030201) {
      return "BIG_ENDIAN";
    } else if (value == 0x03040102) {
      return "PDP_ENDIAN";
    }
  } else if (size == 2) {
    if (value == 0x0102) {
      return "LITTLE_ENDIAN";
    } else if (value == 0x0201) {
      return "BIG_ENDIAN";
    }
  }
  error, "unknown byte order";
}

/*---------------------------------------------------------------------------*/
/* MISCELLANEOUS */

extern is_hash;
/* DOCUMENT is_hash(object)
     Returns 1 if OBJECT is a hash table, else 0.

   SEE ALSO h_new, is_array, is_func, is_list, is_range, is_struct,
            is_stream, is_void. */

extern is_list;
/* DOCUMENT is_list(object)
     Returns 1 if OBJECT is a list, else 0.

   SEE ALSO _lst, is_array, is_hash, is_func, is_range, is_struct,
            is_stream, is_void. */

extern typeIDof;
/* DOCUMENT typeIDof(object)
     Returns type identifier of OBJECT as a long integer:
        0 (T_CHAR)      for an array of char('s)
        1 (T_SHORT)     for an array of short('s)
        2 (T_INT)       for an array of int('s)
        3 (T_LONG)      for an array of long('s)
        4 (T_FLOAT)     for an array of float('s)
        5 (T_DOUBLE)    for an array of double('s)
        6 (T_COMPLEX)   for an array of complex('s)
        7 (T_STRING)    for an array of string('s)
        8 (T_POINTER)   for an array of pointer('s)
        9 (T_STRUCT)    for a structure object
       10 (T_RANGE)     for a range object
       11 (T_LVALUE)    for a lvalue
       12 (T_VOID)      for a void (undefined) object
       13 (T_FUNCTION)  for a function array
       14 (T_BUILTIN)   for a builtin array
       15 (T_STRUCTDEF) for a data type or structure definition
       16 (T_STREAM)    for a file stream
       17 (T_OPAQUE)    for an opaque object

   SEE ALSO typeof, structof. */

extern nrefsof;
/* DOCUMENT nrefsof(object)
     Returns number of references on OBJECT.

   SEE ALSO: unref. */

extern get_encoding;
/* DOCUMENT get_encoding(name);
     Return the data layout for machine NAME, one of:
       "native"   the current machine
          (little-endians)
       "i86"      Intel x86 Linux
       "ibmpc"    IBM PC (2 byte int)
       "alpha"    Compaq alpha
       "dec"      DEC workstation (MIPS), Intel x86 Windows
       "vax"      DEC VAX (H-double)
       "vaxg"     DEC VAX (G-double)
          (big-endians)
       "xdr"      External Data Representation
       "sun"      Sun, HP, SGI, IBM-RS6000, MIPS 32 bit
       "sun3"     Sun-2 or Sun-3 (old)
       "sgi64"    SGI, Sun, HP, IBM-RS6000 64 bit
       "mac"      MacIntosh 68000 (power Mac, Gx are __sun)
       "macl"     MacIntosh 68000 (12 byte double)
       "cray"     Cray XMP, YMP

     The result is a vector of 32 long's as follow:
       [size, align, order] repeated 6  times for  char,  short, int, long,
                            float,  and double, except  that char  align is
                            always  1,   so  result(2)  is   the  structure
                            alignment (see struct_align).
       [sign_address,  exponent_address, exponent_bits,
        mantissa_address, mantissa_bits,
        mantissa_normalization, exponent_bias]  repeated  twice  for  float
                            and double.  See the comment at the top of file
                            prmtyp.i for an explanation of these fields.

     The total number of items is therefore 3*6 + 7*2 = 32.

   SEE ALSO get_primitives, set_primitives, install_encoding. */

func install_encoding(file, encoding)
/* DOCUMENT install_encoding, file, encoding;
     Set layout of  primitive data types for binary  stream FILE.  ENCODING
     may be  one of the  names accepted by  get_encoding or an array  of 32
     integers as explained in get_encoding documentation.

    SEE ALSO: get_encoding, install_struct. */
{
  /* Get encoding parameters with minimal check. */
  if (structof(encoding) == string) {
    p = get_encoding(encoding);
  } else {
    if ((s = structof(encoding)) == long) p = encoding;
    else if (/*s==char || s==short || */s==int) p = long(encoding);
    else error, "bad data type for ENCODING";
    if (numberof(p) != 32) error, "bad number of elements for encoding";
  }

  /* Install primitive definitions. */
  install_struct, file, "char",    1,     1,     p( 3);
  install_struct, file, "short",   p( 4), p( 5), p( 6);
  install_struct, file, "int",     p( 7), p( 8), p( 9);
  install_struct, file, "long",    p(10), p(11), p(12);
  install_struct, file, "float",   p(13), p(14), p(15), p(19:25);
  install_struct, file, "double",  p(16), p(17), p(18), p(26:32);
  struct_align, file, p(2);
}

func same_encoding(a, b)
/* DOCUMENT same_encoding(a, b)
     Compare primitives  A and B  which must be conformable  integer arrays
     with first dimension equals to 32 (see set_primitives).  The result is
     an array  of int's with one  less dimension than A-B  (the first one).
     Some checking is  performed for the operands.  The  byte order for the
     char data type is ignored in the comparison.

   SEE ALSO install_encoding, get_encoding.*/
{
  if (! is_array((d = dimsof(a, b))) || d(1) < 1 || d(2) != 32)
    error, "bad dimensions";
  diff = abs(a - b);
  if ((s = structof(diff)) != long && s != int) error, "bad data type";
  if (anyof(a(1,..) != 1) || anyof(b(1,..) != 1))
    error, "unexpected sizeof(char) != 1";
  diff(3, ..) = 0; /* ignore byte order for type char */
  return ! diff(max,);
}

extern symbol_exists;
/* DOCUMENT symbol_exists(name)
     Check  if variable/function named  NAME exists.   This routine  can be
     used  prior  to  symbol_def  to  check existence  of  a  symbol  since
     symbol_def raise an error for non-existing symbol.

   SEE ALSO symbol_def, symbol_names, symbol_set.*/

extern symbol_names;
/* DOCUMENT symbol_names()
       -or- symbol_names(flags)
     Return an  array of  strings with  the names of  all symbols  of given
     type(s) found in  global symbol table.  To select  the type of symbol,
     FLAGS is be the bitwise-or of one or more of the following bits:
         1 - basic array symbols
         2 - structure instance symbols
         4 - range symbols
         8 - nil symbols (i.e. symbols undefined at current scope level)
        16 - interpreted function symbols
        32 - builtin function symbols
        64 - structure definition symbols
       128 - file stream symbols
       256 - opaque symbols

     The  special value  FLAGS=-1 can  be used  to get  all names  found in
     global symbol table.   The default (if FLAGS is nil  or omitted) is to
     return the names of all symbols but the nil ones.

   SEE ALSO symbol_def, symbol_exists, symbol_info, symbol_set.*/

func symbol_info(____n____)
/* DOCUMENT symbol_info, flags;
       -or- symbol_info, names;
       -or- symbol_info;
     Print out some information about  Yorick's symbols.  FLAGS is a scalar
     integer used to select symbol types (as in symbol_names).  NAMES is an
     array  of symbol  names.  If  argument  is omitted  or undefined,  all
     defined array symbols get selected (as with FLAGS=3).

   SEE ALSO: memory_info, symbol_def, symbol_names.*/
{
  /* attempt to use _very_ odd names to avoid clash with caller */
  if (is_void(____n____)) ____n____ = symbol_names(3);
  else if (structof(____n____) != string) ____n____ = symbol_names(____n____);
  for (____i____=1 ; ____i____<=numberof(____n____) ; ++____i____) {
    write, format="%s:", ____n____(____i____);
    info, symbol_def(____n____(____i____));
  }
}

func memory_info(____a____)
/* DOCUMENT memory_info;
       -or- memory_info, count;
     Print out some information about memory occupation. If COUNT is
     specified, the COUNT biggest (in bytes) symbols are listed (use
     COUNT<0 to list all symbols sorted by size).

   BUGS:
     Only the memory used by Yorick's array symbols is considered,
     e.g. arrays only referenced by a pointer are not considered.  Symbols
     which are aliases (e.g. by using eq_nocopy) may be considered several
     times.

   SEE ALSO: symbol_def, symbol_info, symbol_names. */
{
  ____n____ = symbol_names(3);
  ____i____ = numberof(____n____);
  ____s____ = array(long, ____i____);
  while (____i____ > 0) {
    ____s____(____i____) = sizeof(symbol_def(____n____(____i____)));
    --____i____;
  }
  ____i____ = sum(____s____);
  write, format="Total memory used by array symbols: %d bytes (%.3f Mb)\n",
    ____i____, ____i____/1024.0^2;
  if (____a____) {
    ____i____ = sort(____s____);
    if (____a____ > 0 && ____a____ < numberof(____i____))
      ____i____ = ____i____(1-____a____:0);
    ____n____ = ____n____(____i____);
    ____s____ = ____s____(____i____);
    ____i____ = numberof(____i____);
    if (____i____ > 1) {
      write, format="The %d biggest symbols are:\n", ____i____;
    } else {
      write, format="%s", "The biggest symbol is:\n";
    }
    ____a____ = swrite(format="  %%%ds: %%%.0fd bytes,",
                       max(strlen(____n____)), ceil(log10(max(____s____))));
    while (____i____ > 0) {
      write, format=____a____, ____n____(____i____), ____s____(____i____);
      info, symbol_def(____n____(____i____));
      --____i____;
    }    
  }
}

func about(____n____, ____a____)
/* DOCUMENT about, pattern;
       -or- about, pattern, 1;
     Search and display documentation about functions (or all symbols if
     second argument is true) matching regular expression PATTERN.  If
     multiple matches are found, the user is prompted to select a subject.
     PATTERN may be a string, a compiled regular expression or a function
     or structure definition.  If PATTERN is a string with a trailing "/i",
     the other part of the regular expression is interpreted so as to
     ignore case.

   SEE ALSO help, info, symbol_def, symbol_names, regmatch,
            select_item_in_string_list. */
{
  /* attempt to use _very_ odd names to avoid clash with caller */
  ____a____ = symbol_names((____a____ ? -1 : 48)); /* get names of symbols */
  if (structof(____n____) == string) {
    if (strpart(____n____, -1:0) == "/i") {
      ____n____ = regcomp(strpart(____n____, 1:-2), icase=1);
    }
  } else if (typeof(____n____) != "regex") {
    ____n____ = nameof(____n____);
    if (structof(____n____) != string)
      error, "expecting a string, a function, or a structure definition";
  }
  ____n____ = where(regmatch(____n____, ____a____));
  if (! is_array(____n____)) {
    write, " Sorry no match found.";
    return;
  }
  ____a____ = ____a____(____n____);
  ____a____ = ____a____(sort(____a____));
  ____a____ = select_item_in_string_list(____a____, bol=" ",
                                         prompt=" Choose one subject: ");
  if (! is_void(____a____)) help, symbol_def(____a____);
}

extern swap;
/* DOCUMENT swap, a, b;
     Exchanges  the contents  of variables  A and  B without  requiring any
     temporary copy.  The result of the call is identical to:
       tmp = a; a = b; b = tmp;
     which makes  a copy of  A and then  a copy of B.   Another possibility
     which avoids any copy of A nor B is:
       local tmp;
       eq_nocopy, tmp, a; eq_nocopy, a, b; eq_nocopy, b, tmp;

   SEE ALSO: eq_nocopy, unref. */

extern unref;
/* DOCUMENT unref(x)
     Returns X,  destroying X in the process  if it is an  array (useful to
     deal with temporary big arrays).  Written after Yorick's FAQ.

   SEE ALSO: eq_nocopy, swap, nrefsof. */

extern current_include;
/* DOCUMENT current_include()
     If Yorick is parsing a file, this function returns the absolute path
     of this file; otherwise, this function returns nil.

   SEE ALSO include, require. */

extern get_includes;
/* DOCUMENT get_includes()
     Returns an array of strings with the names of all included files so
     far.

   SEE ALSO: set_path, current_include. */

extern get_path;
/* DOCUMENT get_path()
     Returns an array of strings with the current include file directories.

   SEE ALSO: set_path. */

/*---------------------------------------------------------------------------*/
/* QUERY ROUTINES */

local __select_file_dir;
func select_file(dir, prompt=, width=, forever=, all=, suffix=, pattern=)
/* DOCUMENT select_file()
       -or- select_file(dir)
     Interactively  select name  of an  existing file  starting  at current
     working  directory or at  last selected  directory or  at DIR  if this
     argument  is specified.  The  function returns  full path  of selected
     file or nil  [] if no valid selection is made.   If keyword FOREVER is
     true, a file must be selected for the function to return.

     If keyword ALL  is true, then all files  and directories get displayed
     -- even the "hidden" ones which  name start with a dot.  In any cases,
     the current  and parent  directories ("." and  "..") get  displayed to
     allow the  user to  re-scan the  current directory or  to go  into the
     parent directory.

     Keyword SUFFIX  can be set  to a scalar  string to only  display files
     that match SUFFIX.

     Keyword  PATTERN can be  set to  a regular  expression to  select only
     files that match PATTERN.  For instance,  PATTERN="\\.(tgz|tar\\.gz)$"
     would match any files with suffix ".tgz" or ".tar.gz".

     Keyword WIDTH can  be used to specify a different  text width than the
     default of 79 characters.

     Keyword PROMPT can be set to change the default prompt:
       " Select file/directory: "

   SEE ALSO lsdir, regmatch, pretty_print_string_list. */
{
  /* fool codger */ extern __select_file_dir;
  local dir_list;
  if (is_void(width)) width = 79;
  if (is_void(prompt)) prompt=" Select file/directory: ";
  if (structof(pattern) == string) pattern = regcomp(pattern, nosub=1);
  cwd = get_cwd();
  if (! is_void(dir)) __select_file_dir = dir;
  if (structof(__select_file_dir) != string) {
    __select_file_dir = cwd;
  } else {
    __select_file_dir = cd(__select_file_dir);
  }

  suffix_length = is_void(suffix) ? 0 : strlen(suffix);

  hline = "-------------------------------------";
  for (;;) {
    file_list = lsdir(__select_file_dir, dir_list);
    if (! all) {
      if ((n = numberof(file_list))) {
        i = where(strpart(file_list, 1:1) != ".");
        if (numberof(i) != n) file_list = file_list(i);
      }
      if ((n = numberof(dir_list))) {
        i = where(strpart(dir_list, 1:1) != ".");
        if (numberof(i) != n) dir_list = dir_list(i);
      }
    }
    if (suffix_length > 0 && (n = numberof(file_list))) {
      i = where(strpart(file_list, 1-suffix_length:0) == suffix);
      if (numberof(i) != n) file_list = file_list(i);
    }
    if (pattern && (n = numberof(file_list))) {
      i = where(regmatch(pattern, file_list));
      if (numberof(i) != n) file_list = file_list(i);
    }
    grow, dir_list, ".", ".."; /* use . to allow reading directory again */
    dir_list = dir_list(sort(dir_list));
    list = dir_list + "/";
    if (is_array(file_list)) {
      grow, list, file_list(sort(file_list));
      file_list = [];
    }
    ndirs = numberof(dir_list);
    number = numberof(list);

    /* Print out directory list. */
    text = pretty_print_string_list(list, numbered=": ", width=width,
                                    sep=, eol=" ", bol="| ", maxcols=);
    text_len = strlen(text(1))+1;
    len = max(width, text_len);
    while (strlen(hline) < len) hline += hline;
    head_line = "[" + __select_file_dir + "]";
    n = (len - strlen(head_line) - 2)/2;
    if (n > 0) head_line = strpart(hline, 1:n)+head_line;
    n = len - strlen(head_line) - 2;
    if (n > 0) head_line += strpart(hline, 1:n);
    write, format=",%s.\n", head_line;
    write, format=swrite(format="%%-%ds|\n", len-1), text;
    write, format="`%s'\n", strpart(hline, 1:len-2);
    for (;;) {
      t = string(0);
      k = 0;
      s = rdline(prompt=prompt);
      if (sread(s, format="%d %s", k, t) == 1 && k >= 1 && k <= number) break;
      if (numberof((k = where(list == s))) == 1) {
        k = k(1);
        break;
      }
      if (numberof((k = where(dir_list == s))) == 1) {
        k = k(1);
        break;
      }
      if (! forever) {
        cd, cwd;
        return;
      }
    }
    if (k > ndirs) {
      cd, cwd;
      return __select_file_dir + list(k);
    }
    __select_file_dir = cd(__select_file_dir + dir_list(k));
  }
}

func select_item_in_string_list(list, index=,  prompt=, forever=,
                                numfmt=, width=, sep=, eol=, bol=, maxcols=)
/* DOCUMENT select_item_in_string_list(list)
     Print out  array of strings LIST  (using pretty_print_string_list) and
     interactively ask  the user a number/item  in the list  and return the
     selected item.  If keyword INDEX  is true, the item number is returned
     rather  than its value.   The prompt  string can  be set  with keyword
     PROMPT (default is " Select one  item: ").  If keyword FOREVER is true
     the user is prompted until a valid choice is mage.

     Other  keywords  are passed  to  pretty_print_string_list: NUMFMT  (as
     NUMBERED), WIDTH, SEP, EOL, BOL and MAXCOLS.

   SEE ALSO pretty_print_string_list. */
{
  number = numberof(list);
  pretty_print_string_list, list, numbered=(is_void(numfmt) ? " - " : numfmt),
    width=width, sep=sep, eol=eol, bol=bol, maxcols=maxcols;
  if (is_void(prompt)) prompt=" Select one item: ";
  for (;;) {
    t = string(0);
    k = 0;
    s = rdline(prompt=prompt);
    if (sread(s, format="%d %s", k, t) == 1 && k >= 1 && k <= number) break;
    if (numberof((k = where(list == s))) == 1) {
      k = k(1);
      break;
    }
    if (! forever) return;
  }
  return (index ? k : list(k));
}

func pretty_print_string_list(list, numbered=, width=, start=,
                              sep=, eol=, bol=, maxcols=)
/* DOCUMENT pretty_print_string_list, list;
       -or- pretty_print_string_list(list);
     Write  array of  strings LIST  in  columns.  In  subroutine form,  the
     result is  printed to standard  output; otherwise, return an  array of
     strings (one per row).

     The  maximum  width (in  number  of characters)  of  each  row can  be
     precised with  keyword WIDTH  (default 79).  But  actual width  may be
     larger, since at least one column is produced.

     The maximum number of columns  may be limited by using keyword MAXCOLS
     (by default, there is no limit).

     Keywords BOL,  SEP and  EOL, can be  set to  scalar strings to  use at
     begin-of-line, between each  column, at end-of-line respectively.  SEP
     can  also be  the number  of spaces  to insert  between  columns.  The
     default are: BOL="", SEP=5 (five spaces) and EOL=string(0).

     Keyword NUMBERED can be used to  number items. If NUMBERED is a scalar
     string  which  contains  a "%d",  it  is  used  to format  the  index.
     Otherwise, NUMBERED  is the string or  the number of spaces  to use as
     separator between indices and items.  For instance:
       numbered="[%d] "  yields: "[1] first_item    [2] second_item  ..."
       numbered=" - "    yields: "1 - first_item    2 - second_item  ..."
       numbered=3        yields: "1 first_item      2 second_item ..."
     Keyword START can be used to specify the starting  index for numbering
     items (default START=1).

   SEE ALSO swrite. */
{
  number = numberof(list);
  if (numbered) {
    /* Numbered list. */
    if (is_void(start)) start = 1;
    index = indgen(start:start+number-1);
    if (start >= 0) ndigits = 1 + long(floor(log10(index(0))));
    else ndigits = 2 + long(floor(log10(-start)));
    if (structof(numbered) != string) {
      /* Convert integer into spaces. */
      index = swrite(format=swrite(format="%%%dd%%%ds", ndigits, numbered),
                     index, "");
    } else if (dimsof(numbered)(1)==0 && strmatch(numbered, "%d")) {
      index = swrite(format=numbered, index);
      len = strlen(index);
      if (max(len) != min(len)) {
        /* Justify index list. */
        index = swrite(format=swrite(format="%%%ds", max(len)), index);
      }
    } else {
      /* Use NUMBERED as separator. */
      index = swrite(format=swrite(format="%%%dd%s", ndigits),
                     index) + numbered;
    }
    list = index + list(*);
  } else {
    list = list(*);
  }

  if (is_void(bol)) bol = "";
  if (is_void(eol)) eol = string(0);
  if (is_void(width)) width = 79;
  if (structof(sep) != string) {
    /* Convert margin separator into spaces. */
    if (is_void(sep)) sep = 5;
    sep = swrite(format=swrite(format="%%%ds", sep), "");
  }
  len = max(strlen(list));
  slen = strlen(sep);
  ncols = (width + slen - strlen(bol) - strlen(eol))/(len + slen);
  if (! is_void(maxcols) && ncols > maxcols) ncols = maxcols;
  if (ncols < 1) ncols = 1;
  nrows = (number+ncols-1)/ncols;

  (tmp = array(string, nrows, ncols))(1:number) = unref(list);
  if (ncols > 1) {
    fmt = swrite(format="%%-%ds%s", len, sep);
    for (j=1 ; j<ncols ; ++j) bol += swrite(format=fmt, tmp(,j));
  }
  fmt = (eol ? swrite(format="%%-%ds%s", len, eol) : "%s");
  bol += swrite(format=fmt, tmp(,ncols));
  if (! am_subroutine()) return bol;
  write, format="%s\n", bol;
}

/*---------------------------------------------------------------------------*/
