(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     18668,        480]*)
(*NotebookOutlinePosition[     19304,        503]*)
(*  CellTagsIndexPosition[     19260,        499]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(lm[x_, y_] := \ 
      q0\  - \ \((q1*\((Cos[a]*\((x - x0)\)\  - \ Sin[a]*\((y - y0)\))\))\)^2
           - \ \((q2*\((Sin[a]*\((x - x0)\)\  + \ Cos[a]*\((y - y0)\))\))\)^2
          \)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[{
        a1\  == \ 
          q0 - \((q1\^2\ x0\^2 + q2\^2\ y0\^2)\)\ Cos[a]\^2 - 
            \((q2\^2\ x0\^2 + q1\^2\ y0\^2)\)\ Sin[a]\^2 + 
            \((q1 - q2)\)\ \((q1 + q2)\)\ x0\ y0\ Sin[2\ a], \n\ \ \ \ \ \ \ 
        a2\  == \ 
          2\ \((q1\^2\ x0\ Cos[a]\^2 - 
                \((q1\^2 - q2\^2)\)\ y0\ Cos[a]\ Sin[a] + 
                q2\^2\ x0\ Sin[a]\^2)\), \n\ \ \ \ \ \ \ 
        a3\  == 2\ 
            \((q2\^2\ y0\ Cos[a]\^2 - 
                \((q1\^2 - q2\^2)\)\ x0\ Cos[a]\ Sin[a] + 
                q1\^2\ y0\ Sin[a]\^2)\), \n\ \ \ \ \ \ \ 
        a4\  == \ \(-q1\^2\)\ Cos[a]\^2 - q2\^2\ Sin[a]\^2, \n\ \ \ \ \ \ \ 
        a5\  == \ \((q1\^2 - q2\^2)\)\ Sin[2\ a], \n\ \ \ \ \ \ \ 
        a6\  == \ \(-q2\^2\)\ Cos[a]\^2 - q1\^2\ Sin[a]\^2}, 
      \ {q0, \ \ q1, \ \ q2, \ \ x0, \ \ y0, \ \ a}]\)], "Input"],

Cell[BoxData[
    \(Solve::"ifun" \( : \ \) 
      "Inverse functions are being used by \!\(Solve\), so some solutions may \
not be found."\)], "Message"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \({q0 \[Rule] 
        \(a3\^2\ a4 - a2\ a3\ a5 + a2\^2\ a6 + 
            a1\ \((a5\^2 - 4\ a4\ a6)\)\)\/\(a5\^2 - 4\ a4\ a6\), 
      y0 \[Rule] \(\(-2\)\ a3\ a4 + a2\ a5\)\/\(\(-a5\^2\) + 4\ a4\ a6\), 
      x0 \[Rule] \(\(-a3\)\ a5 + 2\ a2\ a6\)\/\(a5\^2 - 4\ a4\ a6\), 
      q1 \[Rule] 
        \(-\(\@\(\(-a4\) - \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2\)\), 
      q2 \[Rule] 
        \(-\(\@\(\(-a4\) + \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2\)\), 
      a \[Rule] 
        \(-ArcCos[
            \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                    \)\)]\)}\)], "Input"],

Cell[BoxData[
    \({q0 \[Rule] 
        \(a3\^2\ a4 - a2\ a3\ a5 + a2\^2\ a6 + 
            a1\ \((a5\^2 - 4\ a4\ a6)\)\)\/\(a5\^2 - 4\ a4\ a6\), 
      y0 \[Rule] \(\(-2\)\ a3\ a4 + a2\ a5\)\/\(\(-a5\^2\) + 4\ a4\ a6\), 
      x0 \[Rule] \(\(-a3\)\ a5 + 2\ a2\ a6\)\/\(a5\^2 - 4\ a4\ a6\), 
      q1 \[Rule] 
        \(-\(\@\(\(-a4\) - \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2\)\), 
      q2 \[Rule] 
        \(-\(\@\(\(-a4\) + \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2\)\), 
      a \[Rule] 
        \(-ArcCos[
            \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                    \)\)]\)}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[
      \(a3\^2\ a4 - a2\ a3\ a5 + a2\^2\ a6 + 
          a1\ \((a5\^2 - 4\ a4\ a6)\)\)\/\(a5\^2 - 4\ a4\ a6\)]\)], "Input"],

Cell["\<\
(a3**2*a4 - a2*a3*a5 + a2**2*a6 + a1*(a5**2 - 4*a4*a6))/(a5**2 - \
4*a4*a6)\
\>", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[\(\(-2\)\ a3\ a4 + a2\ a5\)\/\(\(-a5\^2\) + 4\ a4\ a6\)]\)],
   "Input"],

Cell["(-2*a3*a4 + a2*a5)/(-a5**2 + 4*a4*a6)", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[\(\(-a3\)\ a5 + 2\ a2\ a6\)\/\(a5\^2 - 4\ a4\ a6\)]\)], 
  "Input"],

Cell["(-(a3*a5) + 2*a2*a6)/(a5**2 - 4*a4*a6)", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[\@\(\(-a4\) - \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2]
      \)], "Input"],

Cell["Sqrt(-a4 - Sqrt(a5**2 + (a4 - a6)**2) - a6)/Sqrt(2)", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[\@\(\(-a4\) + \@\(a5\^2 + \((a4 - a6)\)\^2\) - a6\)\/\@2]
      \)], "Input"],

Cell["Sqrt(-a4 + Sqrt(a5**2 + (a4 - a6)**2) - a6)/Sqrt(2)", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FortranForm[
      \(-ArcCos[
          \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2\)\)]
          \)]\)], "Input"],

Cell["\<\
-ArcCos(-(Sqrt(1 + (a4 - a6)/Sqrt(a5**2 + (a4 - \
a6)**2))/Sqrt(2)))\
\>", "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \({Cos[
        \(-ArcCos[
            \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                    \)\)]\)], \ 
      Sin[\(-ArcCos[
            \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                    \)\)]\)]}\)], "Input"],

Cell[BoxData[
    \({\(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2\)\), 
      \(-\@\(1 + 
            1\/2\ \((\(-1\) - \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\))
                \)\)\)}\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \({\(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2\)\), 
      \(-\(\@\(1 + \(\(-a4\) + a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
              \)\)}\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({FortranForm[
        Sin[a] = 
          \(-\(\@\(1 + \(\(-a4\) + a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                  \)\)], FortranForm[
        Cos[a] = 
          \(-\(\@\(1 + \(a4 - a6\)\/\@\(a5\^2 + \((a4 - a6)\)\^2\)\)\/\@2
                  \)\)]}\)], "Input"],

Cell[BoxData[
    \(Set::"write" \( : \ \) "Tag \!\(Sin\) in \!\(Sin[a]\) is Protected."\)],
   "Message"],

Cell[BoxData[
    \(Set::"write" \( : \ \) "Tag \!\(Cos\) in \!\(Cos[a]\) is Protected."\)],
   "Message"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[
          
          StyleBox[
            \(-\((Sqrt 
                  \((1\  + \ 
                        \((\(-a4\)\  + \ a6)\)/Sqrt 
                          \((a5**2\  + \ \((a4\  - \ a6)\)**2)\))\)/Sqrt 
                  \((2)\))\)\),
            ShowStringCharacters->True,
            NumberMarks->True],
          FortranForm[ 
            Times[ -1, 
              Power[ 2, 
                Rational[ -1, 2]], 
              Power[ 
                Plus[ 1, 
                  Times[ 
                    Power[ 
                      Plus[ 
                        Power[ a5, 2], 
                        Power[ 
                          Plus[ a4, 
                            Times[ -1, a6]], 2]], 
                      Rational[ -1, 2]], 
                    Plus[ 
                      Times[ -1, a4], a6]]], 
                Rational[ 1, 2]]]],
          Editable->True,
          AutoDelete->True], ",", 
        InterpretationBox[
          
          StyleBox[
            \(-\((Sqrt 
                  \((1\  + \ 
                        \((a4\  - \ a6)\)/Sqrt 
                          \((a5**2\  + \ \((a4\  - \ a6)\)**2)\))\)/Sqrt 
                  \((2)\))\)\),
            ShowStringCharacters->True,
            NumberMarks->True],
          FortranForm[ 
            Times[ -1, 
              Power[ 2, 
                Rational[ -1, 2]], 
              Power[ 
                Plus[ 1, 
                  Times[ 
                    Power[ 
                      Plus[ 
                        Power[ a5, 2], 
                        Power[ 
                          Plus[ a4, 
                            Times[ -1, a6]], 2]], 
                      Rational[ -1, 2]], 
                    Plus[ a4, 
                      Times[ -1, a6]]]], 
                Rational[ 1, 2]]]],
          Editable->True,
          AutoDelete->True]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cos[
      ArcTan[\(-\@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\), 
        \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)]]\)], 
  "Input"],

Cell[BoxData[
    \(Cos[
      ArcTan[\(-\@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\), 
        \@\(1 - \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)]]\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(1\/\((c5\^2 - 4\ c4\ c6)\)\^2\((
      c\^4\ \((81\ c4 + 196\ c6 + 
            126\ \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
              \@\(\(c4 - c6 + 
                      \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2 + 
                        c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ 
              \@\(\(\( - c4\) + c6 + 
                      \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2 + 
                        c5\^2 - 2\ c4\ c6 + c6\^2\)\))\) - 
        c\^2\ \((c5\^2 - 4\ c4\ c6)\)\ 
          \((18\ c4\ x + 
              14\ \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
                \@\(\(c4 - c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ 
                \@\(\(\( - c4\) + c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ x + 28\ c6\ y + 
              9\ \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
                \@\(\(c4 - c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ 
                \@\(\(\( - c4\) + c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ y)\) + 
        \((c5\^2 - 4\ c4\ c6)\)\ 
          \((c3\^2\ c4 - c2\ c3\ c5 + c1\ c5\^2 + c2\^2\ c6 - 4\ c1\ c4\ c6 + 
              c4\ c5\^2\ x\^2 - 4\ c4\^2\ c6\ x\^2 + 
              c5\^2\ \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
                \@\(\(c4 - c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ 
                \@\(\(\( - c4\) + c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ x\ y - 
              4\ c4\ \@\(c5\^2 + \((c4 - c6)\)\^2\)\ c6\ 
                \@\(\(c4 - c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ 
                \@\(\(\( - c4\) + c6 + 
                        \@\(c4\^2 + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\/\@\(c4\^2
                             + c5\^2 - 2\ c4\ c6 + c6\^2\)\)\ x\ y + 
              c5\^2\ c6\ y\^2 - 4\ c4\ c6\^2\ y\^2)\))\)\)\)], "Input"],

Cell[BoxData[
    \(c1 + 
      \(\(-c3\^2\)\ c4 + c2\ c3\ c5 - c2\^2\ c6\)\/\(\(-c5\^2\) + 
          4\ c4\ c6\) - 
      1\/2\ \((\(-c4\) - \@\(c5\^2 + \((c4 - c6)\)\^2\) - c6)\)\ 
        \((\(\@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(-\(\(9\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\)\)\) - x)
                    \)\)\/\@2 - 
              \(\@\(1 - \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(14\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\) + y)\)\)\/\@2)
            \)\^2 - 1\/2\ 
        \((\(-c4\) + \@\(c5\^2 + \((c4 - c6)\)\^2\) - c6)\)\ 
        \((\(\@\(1 - \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(9\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\) + x)\)\)\/\@2 - 
              \(\@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(14\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\) + y)\)\)\/\@2)
            \)\^2\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(df[x_, y_] := 
      \((c1 + c2*x + c3*y + c4*x*x + c5*x*y + c6*y*y)\) - 
        \((\((c1\  + \ 
                \((c2*c3*c5\  - \ c3*c3*c4\  - \ c2*c2*c6)\)\ /\ 
                  \((4*c4*c6\  - \ c5*c5)\))\)\  - \ 
            \((\((Sqrt[
                      \((\(-c4\) - c6 - Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/
                        2])\)*\((
                    \(-\((x - \((
                                \((c \((3)\)*c \((5)\)\  - \ 
                                      2*c \((2)\)*c \((6)\))\)\ /\ 
                                  \((4*c4*c6\  - \ c5*c5)\))\))\)\)*
                        Sqrt[\((
                              1\  + \ 
                                \((c4 - c6)\)/
                                  Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/2]\  - 
                      \ \((y - 
                            \((\((c \((2)\)*c \((5)\)\  - \ 
                                    2*c \((3)\)*c \((4)\))\)\ /\ 
                                \((4*c4*c6\  - \ c5*c5)\))\))\)*
                        Sqrt[\((
                              1\  - \ 
                                \((c4 - c6)\)/
                                  Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/2])\))
                \)^2\  - \ 
            \((\((Sqrt[
                      \((\(-c4\) - c6 + Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/
                        2])\)*\((
                    \((x - \((\((c \((3)\)*c \((5)\)\  - \ 
                                    2*c \((2)\)*c \((6)\))\)\ /\ 
                                \((4*c4*c6\  - \ c5*c5)\))\))\)*
                        Sqrt[\((
                              1\  - \ 
                                \((c4 - c6)\)/
                                  Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/2]\  - 
                      \ \((y - 
                            \((\((c \((2)\)*c \((5)\)\  - \ 
                                    2*c \((3)\)*c \((4)\))\)\ /\ 
                                \((4*c4*c6\  - \ c5*c5)\))\))\)*
                        Sqrt[\((
                              1\  + \ 
                                \((c4 - c6)\)/
                                  Sqrt[c5*c5\  + \ \((c4 - c6)\)^2])\)/2])\))
                \)^2)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(1\/\((c5\^2 - 4\ c4\ c6)\)\^2\((
      \(-c\^4\)\ 
          \((81\ c4 + 
              126\ \@\(c5\^2\/\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                \@\(c5\^2 + \((c4 - c6)\)\^2\) + 196\ c6)\) + 
        c\^2\ \((
          9\ \@\(c5\^2\/\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
              \@\(c5\^2 + \((c4 - c6)\)\^2\) + 28\ c6)\)\ 
          \((c5\^2 - 4\ c4\ c6)\)\ x + 
        \((c5\^2 - 4\ c4\ c6)\)\ 
          \((\(-c3\^2\)\ c4 + c2\ c3\ c5 - c2\^2\ c6 + 
              c3\ \((c5\^2 - 4\ c4\ c6)\)\ x)\))\)\)\)], "Input"],

Cell[BoxData[
    \(\(-\(\(\(-c3\^2\)\ c4 + c2\ c3\ c5 - c2\^2\ c6\)\/\(\(-c5\^2\) + 
              4\ c4\ c6\)\)\) + c3\ x + c6\ x\^2 + 
      1\/2\ \((\(-c4\) - \@\(c5\^2 + \((c4 - c6)\)\^2\) - c6)\)\ 
        \((\(-\(\(9\ c\^2\ 
                      \@\(1 + 
                          \(c4 - c6
                              \)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\)\/\(\@2\ 
                      \((\(-c5\^2\) + 4\ c4\ c6)\)\)\)\) - 
              \(\@\(1 - \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(14\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\) + x)\)\)\/\@2)
            \)\^2 + 1\/2\ 
        \((\(-c4\) + \@\(c5\^2 + \((c4 - c6)\)\^2\) - c6)\)\ 
        \((\(9\ c\^2\ 
                  \@\(1 - 
                      \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\)\/\(\@2\ 
                  \((\(-c5\^2\) + 4\ c4\ c6)\)\) - 
              \(\@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
                  \((\(14\ c\^2\)\/\(\(-c5\^2\) + 4\ c4\ c6\) + x)\)\)\/\@2)
            \)\^2\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(1\/\(c5\^2 - 4\ c4\ c6\)\((
      c3\^2\ c4 - c2\ c3\ c5 + c1\ c5\^2 + c2\^2\ c6 - 4\ c1\ c4\ c6 + 
        c4\ c5\^2\ x\^2 - 4\ c4\^2\ c6\ x\^2 - 2\ c4\ c5\^2\ x\ x0 + 
        8\ c4\^2\ c6\ x\ x0 + c4\ c5\^2\ x0\^2 - 4\ c4\^2\ c6\ x0\^2 + 
        c5\^2\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x\ y - 
        4\ c4\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ c6\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x\ y - 
        c5\^2\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x0\ y + 
        4\ c4\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ c6\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x0\ y + 
        c5\^2\ c6\ y\^2 - 4\ c4\ c6\^2\ y\^2 - 
        c5\^2\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x\ y0 + 
        4\ c4\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ c6\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x\ y0 + 
        c5\^2\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x0\ y0
           - 4\ c4\ \@\(1 + \(c4 - c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ 
          \@\(c5\^2 + \((c4 - c6)\)\^2\)\ c6\ 
          \@\(1 + \(\(-c4\) + c6\)\/\@\(c5\^2 + \((c4 - c6)\)\^2\)\)\ x0\ y0
           - 2\ c5\^2\ c6\ y\ y0 + 8\ c4\ c6\^2\ y\ y0 + c5\^2\ c6\ y0\^2 - 
        4\ c4\ c6\^2\ y0\^2)\)\)\)], "Input"]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1152}, {0, 864}},
WindowSize->{811, 600},
WindowMargins->{{151, Automatic}, {Automatic, 37}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 216, 4, 27, "Input"],

Cell[CellGroupData[{
Cell[1950, 57, 871, 17, 129, "Input"],
Cell[2824, 76, 154, 3, 23, "Message"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3015, 84, 634, 13, 170, "Input"],
Cell[3652, 99, 635, 13, 185, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4324, 117, 155, 3, 46, "Input"],
Cell[4482, 122, 101, 3, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4620, 130, 104, 2, 44, "Input"],
Cell[4727, 134, 55, 0, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4819, 139, 99, 2, 44, "Input"],
Cell[4921, 143, 56, 0, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5014, 148, 109, 2, 59, "Input"],
Cell[5126, 152, 69, 0, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5232, 157, 109, 2, 59, "Input"],
Cell[5344, 161, 69, 0, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5450, 166, 155, 4, 71, "Input"],
Cell[5608, 172, 94, 3, 40, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5739, 180, 293, 7, 71, "Input"],
Cell[6035, 189, 221, 4, 80, "Output"]
}, Open  ]],
Cell[6271, 196, 198, 3, 71, "Input"],

Cell[CellGroupData[{
Cell[6494, 203, 298, 7, 71, "Input"],
Cell[6795, 212, 106, 2, 23, "Message"],
Cell[6904, 216, 106, 2, 23, "Message"],
Cell[7013, 220, 1965, 60, 43, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9015, 285, 183, 4, 61, "Input"],
Cell[9201, 291, 176, 3, 62, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9414, 299, 2447, 43, 496, "Input"],
Cell[11864, 344, 926, 16, 236, "Output"]
}, Open  ]],
Cell[12805, 363, 2236, 44, 235, "Input"],

Cell[CellGroupData[{
Cell[15066, 411, 551, 12, 190, "Input"],
Cell[15620, 425, 1036, 19, 236, "Output"]
}, Open  ]],
Cell[16671, 447, 1993, 31, 587, "Input"]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

